"""Quotient structure for a geodesic metric space."""

from abc import ABC

import geomstats.backend as gs
from geomstats.geometry.stratified.point_set import PointSetMetric


class Aligner(ABC):
    """Bundle structure.

    Parameters
    ----------
    total_space : PointSet
        Set with quotient structure.
    align_algo : AlignerAlgorithm
        Algorihtm performing alignment.
    """

    def __init__(self, total_space, align_algo=None):
        self._total_space = total_space
        if align_algo is not None:
            self.align_algo = align_algo

    def align(self, point, base_point):
        """Align point to base point.

        Parameters
        ----------
        point : array-like, shape=[..., *point_shape]
            Point to align.
        base_point : array-like, shape=[..., *point_shape]
            Reference point.

        Returns
        -------
        aligned_point: list, shape = [..., *point_shape]
        """
        if hasattr(self, "align_algo"):
            return self.align_algo.align(point, base_point)

        raise NotImplementedError("`align` is not implemented")


class QuotientMetric(PointSetMetric, ABC):
    """Quotient metric.

    Parameters
    ----------
    space : PointSet
        Set to equip with metric.
    total_space : PointSet
        Set with quotient structure.
    """

    def __init__(self, space, total_space):
        self._total_space = total_space
        super().__init__(space)

    def squared_dist(self, point_a, point_b):
        """Compute distance between two points.

        Parameters
        ----------
        point_a : array-like, shape=[..., *point_shape]
        point_b : array-like, shape=[..., *point_shape]

        Returns
        -------
        distance : array-like, shape=[...]
            Distance between the points.
        """
        aligned_point_b = self._total_space.aligner.align(point_b, point_a)
        return self._total_space.metric.squared_dist(
            point_a,
            aligned_point_b,
        )

    def dist(self, point_a, point_b):
        """Compute distance between two points.

        Parameters
        ----------
        point_a : array-like, shape=[..., *point_shape]
        point_b : array-like, shape=[..., *point_shape]

        Returns
        -------
        distance : array-like, shape=[...]
            Distance between the points.
        """
        return gs.sqrt(self.squared_dist(point_a, point_b))

    def geodesic(self, initial_point, end_point):
        """Compute geodesic between two points.

        Parameters
        ----------
        initial_point : array-like, shape=[..., *point_shape]
            Initial point.
        end_point : array-like, shape=[..., *point_shape]
            End point.

        Returns
        -------
        geodesic : callable
            Geodesic function.
        """
        aligned_end_point = self._total_space.aligner.align(end_point, initial_point)
        return self._total_space.metric.geodesic(
            initial_point=initial_point, end_point=aligned_end_point
        )



import math

import geomstats.backend as gs
from geomstats.test.random import RandomDataGenerator
from geomstats.test.test_case import TestCase
from geomstats.test_cases.geometry.manifold import ManifoldTestCase
from geomstats.vectorization import get_n_points


class SurfacesLocalRandomDataGenerator(RandomDataGenerator):
    def __init__(self, space, point, amplitude=2.0):
        super().__init__(space, amplitude)
        self.point = point

    def _get_deformation(self, n_points):
        dim = self.point.shape[-1]
        dof = self.point.shape[0] - 1

        batch_shape = () if n_points == 1 else (n_points,)

        rand_shape = batch_shape + (dof, dim)

        return (
            gs.concatenate(
                [
                    gs.zeros(batch_shape + (1, dim)),
                    gs.reshape(
                        gs.random.rand(math.prod(rand_shape)),
                        rand_shape,
                    ),
                ],
                axis=-2,
            )
            / self.amplitude
        )

    def random_point(self, n_points=1):
        return self.point + self._get_deformation(n_points)

    def random_tangent_vec(self, point):
        n_points = 1 if point.ndim == 2 else point.shape[0]
        return self._get_deformation(n_points)


class SurfaceTestCase(TestCase):
    def test_face_areas(self, point, expected, atol):
        self.assertAllClose(point.face_areas, expected, atol=atol)

    def test_face_normals(self, point, expected, atol):
        self.assertAllClose(point.face_normals, expected, atol=atol)

    def test_face_centroids(self, point, expected, atol):
        self.assertAllClose(point.face_centroids, expected, atol=atol)


class DiscreteSurfacesTestCase(ManifoldTestCase):
    def test_vertex_areas(self, point, expected, atol):
        """Test vertex_areas.

        Vertex area is the area of all of the triangles who are in contact
        with a specific vertex, according to the formula:
        vertex_areas = 2 * sum_incident_areas / 3.0
        """
        res = self.space.vertex_areas(point)
        self.assertAllClose(res, expected, atol=atol)

    def test_normals(self, point, expected, atol):
        """Test normals.

        Recall that the magnitude of each normal vector is equal to
        the area of the face it is normal to.

        We compare the abs value of each normal vector array because:
        note that the "normals" variable here calculates the normals
        as pointing out of the surface, but the way that normals()
        was constructed makes it so that the normal vector could be
        pointing into the surface or out of the surface, (so it could
        either be positive or negative). Because of this, we make all
        of the normal vectors to the cube positive for testing
        purposes.
        """

        def _check_normal(normal, normal_):
            self.assertTrue(
                gs.any(
                    [
                        gs.allclose(normal, normal_, atol=atol),
                        gs.allclose(normal, -normal_, atol=atol),
                    ]
                )
            )

        res = self.space.normals(point)

        n_points = get_n_points(self.space.point_ndim, point)
        if n_points == 1:
            res = gs.expand_dims(res, axis=0)
            expected = gs.expand_dims(expected, axis=0)

        for res_, expected_ in zip(res, expected):
            for normal, normal_ in zip(res_, expected_):
                _check_normal(normal, normal_)

    def test_surface_one_forms_prod(self, point, expected):
        """Test surface one forms."""
        res = self.space.surface_one_forms(point)

        first_vec = res[..., :, 0, :]
        second_vec = res[..., :, 1, :]
        inner_prods = gs.einsum("...ni,...ni->...n", first_vec, second_vec)

        if gs.ndim(inner_prods) == 1:
            inner_prods = gs.expand_dims(inner_prods, axis=0)
            expected = gs.expand_dims(expected, axis=0)

        for inner_prods_, expected_ in zip(inner_prods, expected):
            self.assertTrue(gs.all([prod in expected_ for prod in inner_prods_]))

    def test_faces_area(self, point, expected, atol):
        """Test faces area."""
        res = self.space.face_areas(point=point)
        self.assertAllClose(res, expected, atol=atol)



"""Collections of wrappers around scikit-learn.

Main goal is to make them compatible with geomstats.
Common issues are point shapes (in geomstats, points may
be represented as n-dim arrays)
and backend support.

For maintainability reasons, we wrap a sklearn object only
when it is strictly necessary.
"""

from sklearn.base import RegressorMixin as _RegressorMixin
from sklearn.decomposition import PCA as _PCA
from sklearn.linear_model import LinearRegression as _LinearRegression
from sklearn.metrics import r2_score

import geomstats.backend as gs


class RegressorMixin(_RegressorMixin):
    """Mixin class for all regression estimators in scikit-learn."""

    def score(self, X, y, sample_weight=None):
        """Return the coefficient of determination of the prediction."""
        y_pred = self.predict(X)
        return r2_score(
            gs.to_numpy(y), gs.to_numpy(y_pred), sample_weight=sample_weight
        )


class PCA(_PCA):
    def __init__(
        self,
        n_components=None,
        *,
        copy=True,
        whiten=False,
        svd_solver="auto",
        tol=0.0,
        iterated_power="auto",
        n_oversamples=10,
        power_iteration_normalizer="auto",
        random_state=None,
    ):
        super().__init__(
            n_components,
            copy=copy,
            whiten=whiten,
            svd_solver=svd_solver,
            tol=tol,
            iterated_power=iterated_power,
            n_oversamples=n_oversamples,
            power_iteration_normalizer=power_iteration_normalizer,
            random_state=random_state,
        )
        self._init_shape = None

    def __repr__(self):
        # to use *args and **kwargs
        return object.__repr__(self)

    def _from_numpy(self):
        """Transform learned attributes in the right backend."""
        self.components_ = gs.from_numpy(self.components_)
        self.explained_variance_ = gs.from_numpy(self.explained_variance_)
        self.explained_variance_ratio_ = gs.from_numpy(self.explained_variance_ratio_)
        self.singular_values_ = gs.from_numpy(self.singular_values_)
        self.mean_ = gs.from_numpy(self.mean_)

    @property
    def reshaped_components_(self):
        if self.components_ is None:
            return None
        return gs.reshape(self.components_, (self.n_components, *self._init_shape[1:]))

    @property
    def reshaped_mean_(self):
        if self.mean_ is None:
            return None

        return gs.reshape(self.mean_, self._init_shape[1:])

    def _reshape(self, x):
        return gs.reshape(x, (x.shape[0], -1))

    def _reshape_X(self, X):
        self._init_shape = X.shape
        return self._reshape(X)

    def fit(self, X, y=None):
        super().fit(self._reshape_X(X))
        self._from_numpy()
        return self

    def fit_transform(self, X, y=None):
        out = super().fit_transform(self._reshape_X(X))
        self._from_numpy()
        return out

    def score_samples(self, X, y=None):
        return super().score_samples(self._reshape(X))

    def score(self, X, y=None):
        return super().score(self._reshape(X))


class LinearRegression(_LinearRegression):
    def __init__(
        self,
        *,
        fit_intercept=True,
        copy_X=True,
        n_jobs=None,
        positive=False,
    ):
        super().__init__(
            fit_intercept=fit_intercept, copy_X=copy_X, n_jobs=n_jobs, positive=positive
        )

        self._init_shape_X = None
        self._init_shape_y = None

    def __repr__(self):
        # to use *args and **kwargs
        return object.__repr__(self)

    def _from_numpy(self):
        """Transform learned attributes in the right backend."""
        self.coef_ = gs.from_numpy(self.coef_)
        self.singular_ = gs.from_numpy(self.singular_)
        self.intercept_ = gs.from_numpy(self.intercept_)

    def _validate_data(self, X, y=None, **kwargs):
        # hack to avoid tensor conversion within validate data
        if y is None:
            return X
        return X, y

    def _reshape(self, x):
        return gs.reshape(x, (x.shape[0], -1))

    def _reshape_X(self, X):
        self._init_shape_X = X.shape
        return self._reshape(X)

    def _reshape_y(self, y):
        self._init_shape_y = y.shape
        return self._reshape(y)

    def _reshape_out(self, out):
        return gs.reshape(out, (out.shape[0], *self._init_shape_y[1:]))

    def fit(self, X, y):
        super().fit(gs.to_numpy(self._reshape_X(X)), y=gs.to_numpy(self._reshape_y(y)))
        self._from_numpy()
        return self

    def predict(self, X):
        return self._reshape_out(super().predict(self._reshape(X)))



"""The vector space of Hermitian matrices.

Lead author: Yann Cabanes.
"""

import logging

import geomstats.backend as gs
from geomstats import algebra_utils as utils
from geomstats.geometry.base import ComplexMatrixVectorSpace
from geomstats.geometry.complex_matrices import ComplexMatrices, ComplexMatricesMetric
from geomstats.geometry.matrices import Matrices


def expmh(mat):
    """Compute the matrix exponential for a Hermitian matrix.

    Parameters
    ----------
    mat : array_like, shape=[..., n, n]
        Symmetric matrix.

    Returns
    -------
    exponential : array_like, shape=[..., n, n]
        Exponential of mat.
    """
    n = mat.shape[-1]
    dim_3_mat = gs.reshape(mat, [-1, n, n])
    expm = apply_func_to_eigvalsh(dim_3_mat, gs.exp)
    return gs.reshape(expm, mat.shape)


def powermh(mat, power):
    """Compute the matrix power for a Hermitian matrix.

    Parameters
    ----------
    mat : array_like, shape=[..., n, n]
        Symmetric matrix with non-negative eigenvalues.
    power : float, list
        Power at which mat will be raised. If a list of powers is passed,
        a list of results will be returned.

    Returns
    -------
    powerm : array_like or list of arrays, shape=[..., n, n]
        Matrix power of mat.
    """
    if isinstance(power, list):
        power_ = [lambda ev, p=p: gs.power(ev, p) for p in power]
    else:

        def power_(ev):
            return gs.power(ev, power)

    return apply_func_to_eigvalsh(mat, power_, check_positive=False)


def apply_func_to_eigvalsh(mat, function, check_positive=False):
    """Apply function to eigenvalues and reconstruct the matrix.

    Parameters
    ----------
    mat : array_like, shape=[..., n, n]
        Hermitian matrix.
    function : callable, list of callables
        Function to apply to eigenvalues. If a list of functions is passed,
        a list of results will be returned.
    check_positive : bool
        Whether to check positivity of the eigenvalues.
        Optional. Default: False.

    Returns
    -------
    mat : array_like, shape=[..., n, n]
        Hermitian matrix.
    """
    eigvals, eigvecs = gs.linalg.eigh(mat)
    if check_positive and gs.any(gs.cast(eigvals, gs.get_default_dtype()) < 0.0):
        try:
            name = function.__name__
        except AttributeError:
            name = function[0].__name__

        logging.warning("Negative eigenvalue encountered in %s", name)

    return_list = True
    if not isinstance(function, list):
        function = [function]
        return_list = False
    reconstruction = []

    if gs.is_complex(mat):
        transp_eigvecs = ComplexMatrices.transconjugate(eigvecs)
    else:
        transp_eigvecs = Matrices.transpose(eigvecs)

    for fun in function:
        eigvals_f = fun(eigvals)
        eigvals_f = utils.from_vector_to_diagonal_matrix(eigvals_f)
        reconstruction.append(Matrices.mul(eigvecs, eigvals_f, transp_eigvecs))
    return reconstruction if return_list else reconstruction[0]


class HermitianMatrices(ComplexMatrixVectorSpace):
    """Class for the vector space of Hermitian matrices of size n.

    Parameters
    ----------
    n : int
        Integer representing the shapes of the matrices: n x n.
    """

    def __init__(self, n, equip=True):
        super().__init__(dim=n * (n + 1) - n, shape=(n, n), equip=equip)
        self.n = n

    @staticmethod
    def default_metric():
        """Metric to equip the space with if equip is True."""
        return ComplexMatricesMetric

    def _create_basis(self):
        """Compute the basis of the vector space of symmetric matrices.

        Returns
        -------
        basis : array-like, shape=[dim, n, n]
        """
        diagonal = []
        real_part = []
        complex_part = []
        for row in gs.arange(self.n):
            for col in gs.arange(row, self.n):
                if row == col:
                    indices = [(row, row)]
                    values = [1.0 + 0j]
                    diagonal.append(
                        gs.array_from_sparse(indices, values, (self.n,) * 2)
                    )
                else:
                    indices = [(row, col), (col, row)]
                    values = [1.0 + 0j, 1.0 + 0j]
                    real_part.append(
                        gs.array_from_sparse(indices, values, (self.n,) * 2)
                    )
                    values = [1j, -1j]
                    complex_part.append(
                        gs.array_from_sparse(indices, values, (self.n,) * 2)
                    )
        return gs.vstack(
            [gs.stack(diagonal), gs.stack(real_part), gs.stack(complex_part)]
        )

    def belongs(self, point, atol=gs.atol):
        """Evaluate if a matrix is Hermitian.

        Parameters
        ----------
        point : array-like, shape=[.., n, n]
            Point to test.
        atol : float
            Tolerance to evaluate equality with the transpose.

        Returns
        -------
        belongs : array-like, shape=[...,]
            Boolean evaluating if point belongs to the space.
        """
        belongs = super().belongs(point)
        if gs.any(belongs):
            is_hermitian = ComplexMatrices.is_hermitian(point, atol)
            return gs.logical_and(belongs, is_hermitian)
        return belongs

    @staticmethod
    def projection(point):
        """Make a matrix Hermitian, by averaging with its transconjugate.

        Parameters
        ----------
        point : array-like, shape=[..., n, n]
            Matrix.

        Returns
        -------
        herm : array-like, shape=[..., n, n]
            Symmetric matrix.
        """
        return ComplexMatrices.to_hermitian(point)

    def random_point(self, n_samples=1, bound=1.0):
        """Sample a Hermitian matrix.

        Points are generated by sampling complex matrices from a uniform distribution
        in a box and averaging with the transconjugate.

        Parameters
        ----------
        n_samples : int
            Number of samples.
            Optional, default: 1.
        bound : float
            Side of hypercube support of the uniform distribution.
            Optional, default: 1.0

        Returns
        -------
        point : array-like, shape=[..., n, n]
           Sample.
        """
        cdtype = gs.get_default_cdtype()
        size = self.shape
        if n_samples != 1:
            size = (n_samples,) + self.shape
        point = gs.cast(
            bound * (gs.random.rand(*size) - 0.5) * 2**0.5,
            dtype=cdtype,
        ) + 1j * gs.cast(
            bound * (gs.random.rand(*size) - 0.5) * 2**0.5,
            dtype=cdtype,
        )
        return ComplexMatrices.to_hermitian(point)

    @staticmethod
    def basis_representation(matrix_representation):
        """Convert a Hermitian matrix into a vector.

        Parameters
        ----------
        matrix_representation : array-like, shape=[..., n, n]
            Matrix.

        Returns
        -------
        basis_representation : array-like, shape=[..., n(n+1)/2]
            Vector.
        """
        diag = Matrices.diagonal(matrix_representation)

        up_triang = gs.triu_to_vec(matrix_representation, k=1)
        real_part = gs.real(up_triang)
        complex_part = gs.imag(up_triang)

        vec = gs.hstack([diag, real_part, complex_part])

        return vec



"""Module exposing the `ComplexMatrices` class.

Lead author: Yann Cabanes.
"""

import geomstats.backend as gs
import geomstats.errors
from geomstats.geometry.base import ComplexMatrixVectorSpace
from geomstats.geometry.hermitian import HermitianMetric
from geomstats.geometry.matrices import Matrices
from geomstats.vectorization import repeat_out


class ComplexMatrices(ComplexMatrixVectorSpace):
    """Class for the space of complex matrices (m, n).

    Parameters
    ----------
    m, n : int
        Integers representing the shapes of the matrices: m x n.
    """

    def __init__(self, m, n, equip=True):
        geomstats.errors.check_integer(n, "n")
        geomstats.errors.check_integer(m, "m")
        super().__init__(dim=m * n * 2, shape=(m, n), equip=equip)
        self.m = m
        self.n = n

    @staticmethod
    def default_metric():
        """Metric to equip the space with if equip is True."""
        return ComplexMatricesMetric

    def belongs(self, point, atol=gs.atol):
        """Check if point belongs to the Matrices space.

        Parameters
        ----------
        point : array-like, shape=[..., m, n]
            Point to be checked.
        atol : float
            Unused here.

        Returns
        -------
        belongs : array-like, shape=[...,]
            Boolean evaluating if point belongs to the Matrices space.
        """
        is_matrix = super().belongs(point, atol=atol)
        belongs = gs.logical_and(is_matrix, gs.is_complex(point))
        return belongs

    @staticmethod
    def transconjugate(mat):
        """Return the transconjugate of matrices.

        Parameters
        ----------
        mat : array-like, shape=[..., n, n]
            Matrix.

        Returns
        -------
        transconjugate : array-like, shape=[..., n, n]
            Transconjugated matrix.
        """
        ndim = gs.ndim(mat)
        axes = list(range(0, ndim))
        axes[-1] = ndim - 2
        axes[-2] = ndim - 1
        return gs.transpose(gs.conj(mat), axes)

    @classmethod
    def is_hermitian(cls, mat, atol=gs.atol):
        """Check if a square matrix is Hermitian.

        Parameters
        ----------
        mat : array-like, shape=[..., n, n]
            Matrix.
        atol : float
            Absolute tolerance.
            Optional, default: backend atol.

        Returns
        -------
        is_herm : array-like, shape=[...,]
            Boolean evaluating if the matrix is symmetric.
        """
        return Matrices.equal(mat, cls.transconjugate(mat), atol)

    @classmethod
    def is_hpd(cls, mat, atol=gs.atol):
        """Check if a square matrix is Hermitian positive definite.

        Parameters
        ----------
        mat : array-like, shape=[..., n, n]
            Matrix.
        atol : float
            Absolute tolerance.
            Optional, default: backend atol.

        Returns
        -------
        is_hpd : array-like, shape=[...,]
            Boolean evaluating if the matrix is Hermitian positive definite.
        """
        return gs.logical_and(cls.is_hermitian(mat, atol), Matrices.is_pd(mat))

    @classmethod
    def is_skew_hermitian(cls, mat, atol=gs.atol):
        """Check if a matrix is skew-Hermitian.

        Parameters
        ----------
        mat : array-like, shape=[..., n, n]
            Matrix.
        atol : float
            Absolute tolerance.
            Optional, default: backend atol.

        Returns
        -------
        is_skew_herm : array-like, shape=[...,]
            Boolean evaluating if the matrix is skew-Hermitian.
        """
        is_square = Matrices.is_square(mat)
        if not is_square:
            is_vectorized = gs.ndim(gs.array(mat)) == 3
            return gs.array([False] * len(mat)) if is_vectorized else False
        return Matrices.equal(mat, -cls.transconjugate(mat), atol)

    @classmethod
    def to_hermitian(cls, mat):
        """Make a matrix Hermitian.

        Make a matrix Hermitian by averaging it
        with its transconjugate.

        Parameters
        ----------
        mat : array-like, shape=[..., n, n]
            Matrix.

        Returns
        -------
        herm : array-like, shape=[..., n, n]
            Hermitian matrix.
        """
        return 1 / 2 * (mat + cls.transconjugate(mat))

    @classmethod
    def to_skew_hermitian(cls, mat):
        """Make a matrix skew-Hermitian.

        Make matrix skew-Hermitian by averaging it
        with minus its transconjugate.

        Parameters
        ----------
        mat : array-like, shape=[..., n, n]
            Matrix.

        Returns
        -------
        skew_sym : array-like, shape=[..., n, n]
            Skew-Hermitian matrix.
        """
        return 1 / 2 * (mat - cls.transconjugate(mat))

    def random_point(self, n_samples=1, bound=1.0):
        """Sample from a uniform distribution in a complex cube.

        Parameters
        ----------
        n_samples : int
            Number of samples.
            Optional, default: 1.
        bound : float
            Bound of the interval in which to sample each entry.
            Optional, default: 1.

        Returns
        -------
        point : array-like, shape=[..., m, n]
            Sample.
        """
        cdtype = gs.get_default_cdtype()

        size = (n_samples,) + self.shape if n_samples != 1 else self.shape
        point = gs.cast(bound * (gs.random.rand(*size) - 0.5), dtype=cdtype)
        point += 1j * gs.cast(bound * (gs.random.rand(*size) - 0.5), dtype=cdtype)
        return point

    @classmethod
    def congruent(cls, mat_1, mat_2):
        r"""Compute the congruent action of mat_2 on mat_1.

        This is :math:`mat\_2 \ mat\_1 \ mat\_2^T`.

        Parameters
        ----------
        mat_1 : array-like, shape=[..., n, n]
            Matrix.
        mat_2 : array-like, shape=[..., n, n]
            Matrix.

        Returns
        -------
        cong : array-like, shape=[..., n, n]
            Result of the congruent action.
        """
        return Matrices.mul(mat_2, mat_1, cls.transconjugate(mat_2))

    @staticmethod
    def frobenius_product(mat_1, mat_2):
        """Compute Frobenius inner-product of two matrices.

        The `einsum` function is used to avoid computing a matrix product. It
        is also faster than using a sum an element-wise product.

        Parameters
        ----------
        mat_1 : array-like, shape=[..., m, n]
            Matrix.
        mat_2 : array-like, shape=[..., m, n]
            Matrix.

        Returns
        -------
        product : array-like, shape=[...,]
            Frobenius inner-product of mat_1 and mat_2
        """
        return gs.einsum("...ij,...ij->...", gs.conj(mat_1), mat_2)


class ComplexMatricesMetric(HermitianMetric):
    """Hermitian metric on complex matrices given by Frobenius inner-product."""

    @staticmethod
    def inner_product(tangent_vec_a, tangent_vec_b, base_point=None):
        """Compute Frobenius inner-product of two tangent vectors.

        Parameters
        ----------
        tangent_vec_a : array-like, shape=[..., m, n]
            Tangent vector.
        tangent_vec_b : array-like, shape=[..., m, n]
            Tangent vector.
        base_point : array-like, shape=[..., m, n]
            Base point.
            Optional, default: None.

        Returns
        -------
        inner_prod : array-like, shape=[...,]
            Frobenius inner-product of tangent_vec_a and tangent_vec_b.
        """
        return ComplexMatrices.frobenius_product(tangent_vec_a, tangent_vec_b)

    def squared_norm(self, vector, base_point=None):
        """Compute the square of the norm of a vector.

        Squared norm of a vector associated to the inner product
        at the tangent space at a base point.

        Parameters
        ----------
        vector : array-like, shape=[..., dim]
            Vector.
        base_point : array-like, shape=[..., dim]
            Base point.
            Optional, default: None.

        Returns
        -------
        sq_norm : array-like, shape=[...,]
            Squared norm.
        """
        sq_norm = self.inner_product(vector, vector, base_point)
        sq_norm = gs.real(sq_norm)
        return sq_norm

    def norm(self, vector, base_point=None):
        """Compute norm of a complex matrix.

        Norm of a matrix associated to the Frobenius inner product.

        Parameters
        ----------
        vector : array-like, shape=[..., dim]
            Vector.
        base_point : array-like, shape=[..., dim]
            Base point.
            Optional, default: None.

        Returns
        -------
        norm : array-like, shape=[...,]
            Norm.
        """
        norm = gs.linalg.norm(vector, axis=(-2, -1))
        return repeat_out(self._space.point_ndim, norm, vector, base_point)



import pytest

import geomstats.backend as gs
from geomstats.test_cases.learning._base import BaseEstimatorTestCase


class ToTangentSpaceTestCase(BaseEstimatorTestCase):
    @pytest.mark.random
    def test_fit_transform_is_tangent(self, n_points, atol):
        X = self.data_generator.random_point(n_points=n_points)

        tangent_vecs = self.estimator.fit_transform(X)
        res = self.estimator.space.is_tangent(
            tangent_vecs, self.estimator.base_point_, atol=atol
        )

        expected = gs.ones(n_points, dtype=bool)
        self.assertAllEqual(res, expected)

    @pytest.mark.random
    def test_inverse_transform_after_transform(self, n_points, atol):
        X = self.data_generator.random_point(n_points=n_points)

        self.estimator.fit(X)
        tangent_vecs = self.estimator.transform(X)
        X_ = self.estimator.inverse_transform(tangent_vecs)

        self.assertAllClose(X_, X, atol=atol)



"""Parameterized curves on any given manifold.

Lead author: Alice Le Brigant.
"""

import logging
import math

from scipy.interpolate import CubicSpline, PchipInterpolator

import geomstats.backend as gs
from geomstats.algebra_utils import from_vector_to_diagonal_matrix
from geomstats.geometry.diffeo import AutodiffDiffeo, Diffeo
from geomstats.geometry.euclidean import Euclidean
from geomstats.geometry.fiber_bundle import AlignerAlgorithm, FiberBundle
from geomstats.geometry.hypersphere import Hypersphere
from geomstats.geometry.landmarks import Landmarks
from geomstats.geometry.manifold import register_quotient
from geomstats.geometry.matrices import Matrices
from geomstats.geometry.nfold_manifold import NFoldManifold, NFoldMetric
from geomstats.geometry.pullback_metric import PullbackDiffeoMetric
from geomstats.geometry.quotient_metric import QuotientMetric
from geomstats.numerics.finite_differences import (
    centered_difference,
    forward_difference,
    second_centered_difference,
)
from geomstats.numerics.interpolation import (
    LinearInterpolator1D,
    UniformUnitIntervalLinearInterpolator,
)
from geomstats.vectorization import check_is_batch, get_batch_shape


def insert_zeros(array, axis=-1, end=False):
    """Insert zeros in a given array.

    Insert zeros while taking care of

    Parameters
    ----------
    array : array-like
    axis : int
        Axis in which insert the zeros.
        Must be given backwards.
    end : bool
        If True, zeros are introduced at the end.

    Returns
    -------
    array_with_zeros : array-like
        Shape in the specified axis increases by one.
    """
    array_ndim = len(array.shape[axis:])

    batch_shape = get_batch_shape(array_ndim, array)

    shape = batch_shape + (1,) + array.shape[len(batch_shape) + 1 :]
    zeros = gs.zeros(shape)

    first, second = (array, zeros) if end else (zeros, array)
    return gs.concatenate((first, second), axis=-array_ndim)


class DiscreteCurves:
    """Space of discrete curves."""

    def __new__(
        self, ambient_dim=2, k_sampling_points=10, starting_at_origin=True, equip=True
    ):
        """Instantiate discrete curves.

        If `starting_at_origin=True`, instantiates discrete curves modulo translations
        eventually equipped with an elastic metric.

        If `starting_at_origin=False`, instantiates discrete curves
        eventually equipped with an L2 metric.

        Parameters
        ----------
        ambient_dim : int
            Dimension of the ambient Euclidean space in which curves take values.
        k_sampling_points : int
            Number of sampling points.
        starting_at_origin : bool
            If rotations are quotiented out.
        equip : bool
            If True, equip space with default metric.
        """
        if starting_at_origin:
            return DiscreteCurvesStartingAtOrigin(
                ambient_dim=ambient_dim,
                k_sampling_points=k_sampling_points,
                equip=equip,
            )

        ambient_manifold = Euclidean(ambient_dim)
        space = Landmarks(
            ambient_manifold=ambient_manifold,
            k_landmarks=k_sampling_points,
            equip=False,
        )
        if equip:
            space.equip_with_metric(L2CurvesMetric)

        return space


class DiscreteCurvesStartingAtOrigin(NFoldManifold):
    r"""Space of discrete curves modulo translations.

    Each individual curve is represented by a 2d-array of shape `[
    k_sampling_points - 1, ambient_dim]`.

    This space corresponds to the space of immersions defined below, i.e. the
    space of smooth functions from an interval I into the ambient Euclidean
    space M, with non-vanishing derivative.

    .. math::
        Imm(I, M)=\{c \in C^{\infty}(I, M) \|c'(s)\|\neq 0 \forall s \in I \},

    where the interval of parameters I is taken to be I = [0, 1]
    without loss of generality.

    Parameters
    ----------
    ambient_dim : int
        Dimension of the ambient Euclidean space in which curves take values.
    k_sampling_points : int
        Number of sampling points.
    equip : bool
        If True, equip space with default metric.
    """

    def __init__(self, ambient_dim=2, k_sampling_points=10, equip=True):
        ambient_manifold = Euclidean(ambient_dim)
        super().__init__(ambient_manifold, k_sampling_points - 1, equip=equip)

        self._sphere = Hypersphere(dim=ambient_dim - 1)
        self._discrete_curves_with_l2 = None

    def new(self, equip=True):
        """Create manifold with same parameters."""
        return DiscreteCurvesStartingAtOrigin(
            ambient_dim=self.ambient_manifold.dim,
            k_sampling_points=self.k_sampling_points,
            equip=equip,
        )

    @property
    def ambient_manifold(self):
        """Manifold in which curves take values."""
        return self.base_manifold

    @property
    def k_sampling_points(self):
        """Number of sampling points for the discrete curves."""
        return self.n_copies + 1

    @property
    def discrete_curves_with_l2(self):
        """Copy of discrete curves with the L^2 metric."""
        if self._discrete_curves_with_l2 is None:
            self._discrete_curves_with_l2 = self.new(equip=False).equip_with_metric(
                L2CurvesMetric
            )
        return self._discrete_curves_with_l2

    @staticmethod
    def default_metric():
        """Metric to equip the space with if equip is True."""
        return SRVMetric

    def insert_origin(self, point):
        """Insert origin as first element of point."""
        return insert_zeros(point, axis=-self.point_ndim)

    def projection(self, point):
        """Project a point from discrete curves.

        Removes translation and origin.

        Parameters
        ----------
        point : array-like, shape=[..., k_sampling_points, ambient_dim]

        Returns
        -------
        proj_point : array-like, shape=[..., k_sampling_points - 1, ambient_dim]
        """
        if point.shape[-2] == self.k_sampling_points - 1:
            return gs.copy(point)

        return (point[..., :, :] - gs.expand_dims(point[..., 0, :], axis=-2))[
            ..., 1:, :
        ]

    def random_point(self, n_samples=1, bound=1.0):
        """Sample random curves.

        Sampling on the sphere to avoid chaotic curves.
        """
        sampling_times = gs.linspace(0.0, 1.0, self.k_sampling_points)

        initial_point = self._sphere.random_point(n_samples)
        initial_tangent_vec = self._sphere.random_tangent_vec(initial_point)

        point = self._sphere.metric.geodesic(
            initial_point, initial_tangent_vec=initial_tangent_vec
        )(sampling_times)
        return self.projection(point)

    def interpolate(self, point):
        """Interpolate between the sampling points of a discrete curve.

        Parameters
        ----------
        point : array-like, shape=[..., k_sampling_points, ambient_dim]
            Discrete curve starting at the origin.

        Returns
        -------
        spline : function
            Cubic spline that interpolates between the sampling points
            of the discrete curve.
        """
        k_sampling_points = self.k_sampling_points
        t_space = gs.linspace(0.0, 1.0, k_sampling_points)
        point_with_origin = insert_zeros(point, axis=-self.point_ndim)
        is_batch = check_is_batch(self.point_ndim, point)

        def interpolating_curve(t):
            if not is_batch:
                return gs.from_numpy(
                    CubicSpline(t_space, point_with_origin, axis=-self.point_ndim)(t)
                )
            return gs.stack(
                [
                    gs.from_numpy(
                        CubicSpline(t_space, point_with_origin_, axis=-self.point_ndim)(
                            t
                        )
                    )
                    for point_with_origin_ in point_with_origin
                ]
            )

        return interpolating_curve

    def length(self, point):
        """Compute the length of a discrete curve.

        This is the integral of the absolute value of the velocity.

        Parameters
        ----------
        point : array-like, shape=[..., k_sampling_points, ambient_dim]
            Discrete curve starting at the origin.

        Returns
        -------
        length : array-like, shape=[..., ]
            Length of the discrete curve.
        """
        point_with_origin = self.insert_origin(point)
        velocity = forward_difference(point_with_origin, axis=-self.point_ndim)
        l2_metric = self.discrete_curves_with_l2.metric
        return l2_metric.norm(velocity, point_with_origin[..., :-1, :])

    def normalize(self, point):
        """Rescale discrete curve to have unit length."""
        return gs.einsum("...ij,...->...ij", point, 1 / self.length(point))


class SRVTransform(Diffeo):
    """SRV transform.

    Diffeomorphism between discrete curves starting at origin with
    `k_sampling_points` and landmarks with `k_sampling_points - 1`.

    Parameters
    ----------
    ambient_manifold : Manifold
        Manifold in which curves take values.
    k_sampling_points : int
        Number of sampling points.

    Notes
    -----
    It is currently only implemented for the Euclidean ambient manifold.
    """

    def __init__(self, ambient_manifold, k_sampling_points):
        self.ambient_manifold = ambient_manifold
        self.k_sampling_points = k_sampling_points

        self._point_ndim = self.ambient_manifold.point_ndim + 1

    def __call__(self, base_point):
        r"""Square Root Velocity Transform (SRVT).

        Compute the square root velocity representation of a curve. The
        velocity is computed using the log map.

        In the case of several curves, an index selection procedure allows to
        get rid of the log between the end point of curve[k, :, :] and the starting
        point of curve[k + 1, :, :].

        .. math::
            Q(c) = c'/ |c'|^{1/2}

        Parameters
        ----------
        base_point : array-like, shape=[..., k_sampling_points - 1, ambient_dim]
            Discrete curve.

        Returns
        -------
        image_point : array-like, shape=[..., k_sampling_points - 1, ambient_dim]
            SRV representation.
        """
        ndim = self._point_ndim
        base_point_with_origin = insert_zeros(base_point, axis=-ndim)

        velocity = forward_difference(base_point_with_origin, axis=-ndim)

        pointwise_velocity_norm = self.ambient_manifold.metric.norm(
            velocity, base_point_with_origin[..., :-1, :]
        )
        return gs.einsum(
            "...ij,...i->...ij", velocity, 1.0 / gs.sqrt(pointwise_velocity_norm)
        )

    def inverse(self, image_point):
        r"""Inverse of the Square Root Velocity Transform (SRVT).

        Retrieve a curve from its square root velocity representation.

        .. math::
            c(s) = c(0) + \int_0^s q(u) |q(u)|du

        with:

        - c the curve that can be retrieved only up to a translation,
        - q the srv representation of the curve,
        - c(0) the starting point of the curve.


        See [Sea2011]_ Section 2.1 for details.

        It performs numerical integration on a manifold.

        Parameters
        ----------
        image_point : array-like, shape=[..., k_sampling_points - 1, ambient_dim]
            SRV representation.

        Returns
        -------
        curve : array-like, shape=[..., k_sampling_points - 1, ambient_dim]
            Discrete curve.
        """
        image_point_norm = self.ambient_manifold.metric.norm(image_point)

        dt = 1 / (self.k_sampling_points - 1)

        pointwise_delta_points = gs.einsum(
            "...,...i->...i", dt * image_point_norm, image_point
        )

        return gs.cumsum(pointwise_delta_points, axis=-2)

    def tangent(self, tangent_vec, base_point=None, image_point=None):
        r"""Differential of the square root velocity transform.

        .. math::
            (h, c) -> dQ_c(h) = |c'|^(-1/2) * (h' - 1/2 * <h',v>v)
            v = c'/|c'|

        Parameters
        ----------
        tangent_vec : array-like, shape=[..., k_sampling_points - 1, ambient_dim]
            Tangent vector to curve.
        base_point : array-like, shape=[..., k_sampling_points - 1, ambient_dim]
            Discrete curve.
        image_point : array-like, shape=[..., k_sampling_points - 1, ambient_dim]
            SRV representation.

        Returns
        -------
        d_srv_vec : array-like, shape=[..., k_sampling_points - 1, ambient_dim,]
            Differential of the square root velocity transform at curve
            evaluated at tangent_vec.
        """
        if base_point is None:
            base_point = self.inverse(image_point)

        ndim = self._point_ndim
        base_point_with_origin = insert_zeros(base_point, axis=-ndim)
        tangent_vec_with_zeros = insert_zeros(tangent_vec, axis=-ndim)

        d_vec = forward_difference(tangent_vec_with_zeros, axis=-ndim)
        velocity_vec = forward_difference(base_point_with_origin, axis=-ndim)

        velocity_norm = self.ambient_manifold.metric.norm(velocity_vec)
        unit_velocity_vec = gs.einsum(
            "...ij,...i->...ij", velocity_vec, 1 / velocity_norm
        )

        pointwise_inner_prod = self.ambient_manifold.metric.inner_product(
            d_vec, unit_velocity_vec, base_point_with_origin[..., :-1, :]
        )
        d_vec_tangential = gs.einsum(
            "...ij,...i->...ij",
            unit_velocity_vec,
            pointwise_inner_prod,
        )
        d_srv_vec = d_vec - 1 / 2 * d_vec_tangential
        d_srv_vec = gs.einsum(
            "...ij,...i->...ij", d_srv_vec, 1 / velocity_norm ** (1 / 2)
        )

        return d_srv_vec

    def inverse_tangent(self, image_tangent_vec, image_point=None, base_point=None):
        r"""Inverse of differential of the square root velocity transform.

        .. math::
            (c, k) -> h, \text{ where } dQ_c(h)=k \text{ and } h' = |c'| * (k + <k,v> v)

        Parameters
        ----------
        image_tangent_vec : array-like, shape=[..., k_sampling_points - 1, ambient_dim]
            Tangent vector to SRV representation.
        image_point : array-like, shape=[..., k_sampling_points - 1, ambient_dim]
            SRV representation.
        base_point : array-like, shape=[..., k_sampling_points - 1, ambient_dim]
            Discrete curve.

        Returns
        -------
        tangent_vec : array-like, shape=[..., ambient_dim]
            Inverse of the differential of the square root velocity transform at
            curve evaluated at tangent_vec.
        """
        if base_point is None:
            base_point = self.inverse(image_point)

        ndim = self._point_ndim
        base_point_with_origin = insert_zeros(base_point, axis=-ndim)

        position = base_point_with_origin[..., :-1, :]
        velocity_vec = forward_difference(base_point_with_origin, axis=-ndim)
        velocity_norm = self.ambient_manifold.metric.norm(velocity_vec, position)
        unit_velocity_vec = gs.einsum(
            "...ij,...i->...ij", velocity_vec, 1 / velocity_norm
        )

        pointwise_inner_prod = self.ambient_manifold.metric.inner_product(
            image_tangent_vec,
            unit_velocity_vec,
            position,
        )
        tangent_vec_tangential = gs.einsum(
            "...ij,...i->...ij", unit_velocity_vec, pointwise_inner_prod
        )
        d_vec = image_tangent_vec + tangent_vec_tangential
        d_vec = gs.einsum("...ij,...i->...ij", d_vec, velocity_norm ** (1 / 2))
        increment = d_vec / (self.k_sampling_points - 1)

        return gs.cumsum(increment, axis=-2)


class FTransform(AutodiffDiffeo):
    r"""FTransform.

    The f_transform is defined on the space of curves
    quotiented by translations, which is identified with the space
    of curves with their first sampling point located at 0:

    .. math::
        curve(0) = (0, 0)

    The f_transform is given by the formula:

    .. math::
        Imm(I, R^2) / R^2 \mapsto C^\infty(I, R^2\backslash\{0\})
        c \mapsto 2b |c'|^{1/2} (\frac{c'}{|c'|})^{a/(2b)}

    where the identification :math:`C = R^2` is used and
    the exponentiation is a complex exponentiation, which can make
    the f_transform not well-defined:

    .. math::
        f(c) = 2b r^{1/2}\exp(i\theta * a/(2b)) * \exp(ik\pi * a/b)

    where (r, theta) is the polar representation of c', and for
    any :math:`k \in Z`.

    Parameters
    ----------
    ambient_manifold : Manifold
        Manifold in which curves take values.
    k_sampling_points : int
        Number of sampling points.
    a : float
        Bending parameter.
    b : float
        Stretching parameter.

    Notes
    -----
    It is currently only implemented for the Euclidean ambient manifold with
    dimension 2.

    f_transform is a bijection if and only if a/2b=1.

    If a/2b is an integer not equal to 1:

    - then f_transform is well-defined but many-to-one.

    If a/2b is not an integer:

    - then f_transform is multivalued,
    - and f_transform takes finitely many values if and only if a 2b is rational.
    """

    def __init__(self, ambient_manifold, k_sampling_points, a=1.0, b=None):
        self._check_ambient_manifold(ambient_manifold)

        self.a = a
        if b is None:
            b = a / 2
        self.b = b
        self.ambient_manifold = ambient_manifold
        self.k_sampling_points = k_sampling_points

        shape = (k_sampling_points - 1,) + self.ambient_manifold.shape
        super().__init__(shape, shape)

    def _check_ambient_manifold(self, ambient_manifold):
        if not (isinstance(ambient_manifold, Euclidean) and ambient_manifold.dim == 2):
            raise NotImplementedError(
                "This transformation is only implemented for planar curves:\n"
                "ambient_manifold must be a plane, but it is:\n"
                f"{ambient_manifold} of dimension {ambient_manifold.dim}."
            )

    def _cartesian_to_polar(self, tangent_vec):
        """Compute polar coordinates of a tangent vector from the cartesian ones.

        This function is an auxiliary function used for the computation
        of the f_transform and its inverse, and is applied to the derivative
        of a curve.

        See [KN2018]_ for details.

        Parameters
        ----------
        tangent_vec : array-like, shape=[..., k, ambient_dim]
            Tangent vector, representing the derivative c' of a discrete curve c.

        Returns
        -------
        polar_tangent_vec : array-like, shape=[..., k, ambient_dim]
        """
        k_sampling_points = tangent_vec.shape[-2]
        norms = self.ambient_manifold.metric.norm(tangent_vec)
        arg_0 = gs.arctan2(tangent_vec[..., 0, 1], tangent_vec[..., 0, 0])
        args = [arg_0]

        for i in range(1, k_sampling_points):
            point, last_point = tangent_vec[..., i, :], tangent_vec[..., i - 1, :]
            prod = self.ambient_manifold.metric.inner_product(point, last_point)
            cosine = prod / (norms[..., i] * norms[..., i - 1])
            angle = gs.arccos(gs.clip(cosine, -1, 1))
            det = gs.linalg.det(gs.stack([last_point, point], axis=-1))
            orientation = gs.sign(det)
            arg = args[-1] + orientation * angle
            args.append(arg)

        args = gs.stack(args, axis=-1)
        return gs.stack([norms, args], axis=-1)

    def _polar_to_cartesian(self, polar_tangent_vec):
        """Compute the cartesian coordinates of a tangent vector from polar ones.

        This function is an auxiliary function used for the computation
        of the f_transform.

        Parameters
        ----------
        polar_tangent_vec : array-like, shape=[..., k, ambient_dim]

        Returns
        -------
        tangent_vec : array-like, shape=[..., k, ambient_dim]
            Tangent vector.
        """
        tangent_vec_x = gs.cos(polar_tangent_vec[..., :, 1])
        tangent_vec_y = gs.sin(polar_tangent_vec[..., :, 1])
        norms = polar_tangent_vec[..., :, 0]
        unit_tangent_vec = gs.stack((tangent_vec_x, tangent_vec_y), axis=-1)

        return norms[..., :, None] * unit_tangent_vec

    def __call__(self, base_point):
        r"""Compute the f_transform of a curve.

        The implementation uses formula (3) from [KN2018]_ , i.e. choses
        the representative corresponding to k = 0.

        Parameters
        ----------
        base_point : array-like, shape=[..., k_sampling_points - 1, ambient_dim]
            Discrete curve.

        Returns
        -------
        image_point : array-like, shape=[..., k_sampling_points - 1, ambient_dim]
            F_transform of the curve.
        """
        coeff = self.k_sampling_points - 1

        base_point_with_origin = insert_zeros(base_point, axis=-self._space_point_ndim)

        velocity = coeff * (
            base_point_with_origin[..., 1:, :] - base_point_with_origin[..., :-1, :]
        )
        polar_velocity = self._cartesian_to_polar(velocity)
        speeds = polar_velocity[..., :, 0]
        args = polar_velocity[..., :, 1]

        f_args = args * (self.a / (2 * self.b))
        f_norms = 2 * self.b * gs.sqrt(speeds)
        f_polar = gs.stack([f_norms, f_args], axis=-1)

        return self._polar_to_cartesian(f_polar)

    def inverse(self, image_point):
        r"""Compute the inverse F_transform of a transformed curve.

        This only works if a / (2b) <= 1.
        See [KN2018]_ for details.

        When the f_transform is many-to-one, one antecedent is chosen.

        Parameters
        ----------
        image_point : array-like, shape=[..., k_sampling_points - 1, ambient_dim]
            F transform representation of a discrete curve.

        Returns
        -------
        point : array-like, shape=[..., k_sampling_points - 1, ambient_dim]
            Curve starting at the origin retrieved from its square-root velocity.
        """
        f_polar = self._cartesian_to_polar(image_point)
        f_norms = f_polar[..., :, 0]
        f_args = f_polar[..., :, 1]

        dt = 1 / (self.k_sampling_points - 1)

        delta_points_x = gs.einsum(
            "...i,...i->...i", dt * f_norms**2, gs.cos(2 * self.b / self.a * f_args)
        )
        delta_points_y = gs.einsum(
            "...i,...i->...i", dt * f_norms**2, gs.sin(2 * self.b / self.a * f_args)
        )

        delta_points = gs.stack((delta_points_x, delta_points_y), axis=-1)

        delta_points = 1 / (4 * self.b**2) * delta_points

        return gs.cumsum(delta_points, axis=-2)


class L2CurvesMetric(NFoldMetric):
    """L2 metric on the space of discrete curves.

    L2 metric on the space of regularly sampled discrete curves
    defined on the unit interval. The inner product between two tangent vectors
    is given by the integral of the ambient space inner product, approximated by
    a left Riemann sum.
    """

    def __init__(self, space):
        super().__init__(
            space=space,
            signature=(math.inf, 0, 0),
        )

    @staticmethod
    def riemann_sum(func):
        r"""Compute the left Riemann sum approximation of the integral.

        Compute the left Riemann sum approximation of the integral of a
        function func defined on the unit interval, given by sample points
        at regularly spaced times

        .. math::
            t_i = i / k, \\
            i = 0, ..., k - 1

        where :math:`k` is the number of landmarks (last time is missing).

        Parameters
        ----------
        func : array-like, shape=[..., k_landmarks]
            Sample points of a function at regularly spaced times.

        Returns
        -------
        riemann_sum : array-like, shape=[..., ]
            Left Riemann sum.
        """
        k_sampling_points_minus_one = func.shape[-1]
        dt = 1 / k_sampling_points_minus_one
        return dt * gs.sum(func, axis=-1)

    def inner_product(self, tangent_vec_a, tangent_vec_b, base_point=None):
        """Compute L2 inner product between two tangent vectors.

        The inner product is the integral of the ambient space inner product,
        approximated by a left Riemann sum.

        Parameters
        ----------
        tangent_vec_a : array-like, shape=[..., k_landmarks, ambient_dim]
            Tangent vector to a curve, i.e. infinitesimal vector field
            along a curve.
        tangent_vec_b : array-like, shape=[..., k_landmarks, ambient_dim]
            Tangent vector to a curve, i.e. infinitesimal vector field
            along a curve.
        base_point : array-like, shape=[..., k_landmarks, ambient_dim]
            Discrete curve defined on the unit interval [0, 1].

        Return
        ------
        inner_prod : array_like, shape=[...]
            L2 inner product between tangent_vec_a and tangent_vec_b.
        """
        inner_products = self.pointwise_inner_product(
            tangent_vec_a, tangent_vec_b, base_point
        )
        return self.riemann_sum(inner_products)


class ElasticMetric(PullbackDiffeoMetric):
    r"""Elastic metric on the space of discrete curves.

    Family of elastic metric parametrized by bending and stretching parameters
    a and b.

    For curves in :math:`\mathbb{R}^2`, it pullbacks the L2 metric by
    the F-transform (see [NK2018]_).

    For curves in :math:`\mathbb{R}^d`, it pullbacks a scaled L2 metric by the SRV
    transform (see [BCKKNP2022_). It only works for ratio :math:`a / 2b = 1`.

    When a=1, and b=1/2, it is equivalent to the SRV metric.

    Parameters
    ----------
    a : float
        Bending parameter.
    b : float
        Stretching parameter.

    References
    ----------
    .. [KN2018] S. Kurtek and T. Needham,
        "Simplifying transforms for general elastic metrics on the space of
        plane curves", arXiv:1803.10894 [math.DG], 29 Mar 2018.
    .. [BCKKNP2022] Martin Bauer, Nicolas Charon, Eric Klassen, Sebastian Kurtek,
        Tom Needham, and Thomas Pierron.
        “Elastic Metrics on Spaces of Euclidean Curves: Theory and Algorithms.”
        arXiv, September 20, 2022. https://doi.org/10.48550/arXiv.2209.09862.
    """

    def __init__(
        self,
        space,
        a,
        b=None,
    ):
        if b is None:
            b = a / 2

        self.a = a
        self.b = b

        ambient_manifold = space.ambient_manifold
        k_sampling_points = space.k_sampling_points

        image_space = Landmarks(
            ambient_manifold=space.ambient_manifold,
            k_landmarks=k_sampling_points - 1,
            equip=False,
        )
        image_space.equip_with_metric(L2CurvesMetric)

        if ambient_manifold.dim == 2:
            diffeo = FTransform(space.ambient_manifold, k_sampling_points, a, b)
        elif gs.abs(a / (2 * b) - 1.0) < gs.atol:
            diffeo = SRVTransform(ambient_manifold, k_sampling_points)
            if gs.abs(b - 0.5) > gs.atol:
                scale = 4 * b**2
                image_space.equip_with_metric(scale * image_space.metric)
        else:
            raise ValueError(
                "Cannot instantiate elastic metric for ambient dim="
                f"{ambient_manifold.dim}, a={a} and b={b}. "
                "Ratio must be 1 or ambient_dim=2."
            )

        super().__init__(
            space=space,
            diffeo=diffeo,
            image_space=image_space,
            signature=(math.inf, 0, 0),
        )


class SRVMetric(PullbackDiffeoMetric):
    """Square Root Velocity metric on the space of discrete curves.

    The SRV metric is equivalent to the elastic metric chosen with

    - bending parameter a = 1,
    - stretching parameter b = 1/2.

    It can be obtained as the pullback of the L2 metric by the Square Root
    Velocity Function.

    See [Sea2011]_ for details.

    References
    ----------
    .. [Sea2011] A. Srivastava, E. Klassen, S. H. Joshi and I. H. Jermyn,
        "Shape Analysis of Elastic Curves in Euclidean Spaces,"
        in IEEE Transactions on Pattern Analysis and Machine Intelligence,
        vol. 33, no. 7, pp. 1415-1428, July 2011.
    """

    def __init__(self, space):
        self.a = 1.0
        self.b = 1 / 2
        self._check_ambient_manifold(space.ambient_manifold)

        image_space = self._instantiate_image_space(space)

        diffeo = SRVTransform(space.ambient_manifold, space.k_sampling_points)
        super().__init__(space=space, diffeo=diffeo, image_space=image_space)

    def _check_ambient_manifold(self, ambient_manifold):
        if not isinstance(ambient_manifold, Euclidean):
            raise AssertionError(
                "This metric is only "
                "implemented for discrete curves embedded "
                "in a Euclidean space."
            )

    def _instantiate_image_space(self, space):
        image_space = Landmarks(
            ambient_manifold=space.ambient_manifold,
            k_landmarks=space.k_sampling_points - 1,
            equip=False,
        )
        image_space.equip_with_metric(L2CurvesMetric)
        return image_space


class IterativeHorizontalGeodesicAligner(AlignerAlgorithm):
    r"""Align two curves through iterative horizontal geodesic algorithm.

    This algorithm computes the horizontal geodesic between two curves in the shape
    bundle of curves modulo reparametrizations, and at the same time, aligns the
    end curve with respect to the initial curve. This is done through an iterative
    procedure where the initial curve stays fixed and the sampling points are moved
    on the end curve to obtain its optimal parametrization with respect to the initial
    curve. This procedure is based on the decomposition of any path of curves into a
    horizontal path of curves composed with a path of reparametrizations:
    :math:`c(t, s) = c_h(t, phi(t, s))` where :math:`d/dt c_h(t, .)` is horizontal.
    Here t is the time parameter of the path and s the space parameter of the curves.

    The algorithm sets current_end_curve to be the end curve and iterates three steps:
    1) compute the geodesic between the initial curve and current_end_curve
    2) compute the path of reparametrizations such that the path of its inverses
    transforms this geodesic into a horizontal path of curves
    3) invert this path of reparametrizations to find the horizontal path and update
    current_end_curve to be its end point.
    The algorithm stops when the new current_end_curve is sufficiently
    close to the former current_end_curve.

    Parameters
    ----------
    total_space : Manifold
        Total space with reparametrizations fiber bundle structure.
    n_time_grid : int
        Number of times in which compute the geodesic.
    threshold : float
        When the difference between the new end curve and the current end
        curve becomes lower than this threshold, the algorithm stops.
        Optional, default: 1e-3.
    max_iter : int
        Maximum number of iterations.
        Optional, default: 20.
    tol: float
        Minimal spacing between time samples in the unit segment when
        reparametrizing the end curve.
        Optional, default: 1e-3.
    verbose: boolean
        Optional, default: False.
    save_history : bool
        If True, history is saved in a `self.history`.

    References
    ----------
    .. [LAB2017] A. Le Brigant, "Optimal matching between curves in a manifold",
        in Geometric Science of Information. Springer Lecture Notes in Computer
        Science 10589 (2017), 57 - 64. https://hal.science/hal-04374199.
    """

    def __init__(
        self,
        total_space,
        n_time_grid=100,
        threshold=1e-3,
        max_iter=20,
        tol=1e-3,
        verbose=0,
        save_history=False,
    ):
        super().__init__(total_space)
        self.n_time_grid = n_time_grid
        self.threshold = threshold
        self.max_iter = max_iter
        self.tol = tol
        self.verbose = verbose
        self.save_history = save_history
        if save_history:
            self.history = None

    @staticmethod
    def _euler_step_forward(current, increment, step, tol):
        r"""Perform Euler step while enforcing increasing solution.

        Compute the new state phi(t+dt,.) from the current state phi(t,.) using
        the Euler step: new = current + step * increment, i.e.
        :math:`phi(t+dt,.) = phi(t,.) + dt * d/dt phi(t,.)`
        while ensuring that the result is an increasing function on the unit
        interval that preserves the end points 0 and 1. This is done by alternatively
        computing phi(t+dt,s) and phi(t+dt,1-s), for s increasing from 0 to 0.5,
        in a way that ensures that the spacing with the previous and following
        sampling points are greater than a given tolerance.
        """
        k_sampling_points = current.shape[0]
        max_index = k_sampling_points // 2
        new = gs.copy(current)
        sign = gs.sign(tol)
        for index in range(1, max_index):
            new_index = current[index] + step * increment[index - 1]
            new_index_min = new[index - 1] + tol
            new_index_max = (
                new[k_sampling_points - index] - (k_sampling_points - 2 * index) * tol
            )
            new_index = sign * gs.maximum(sign * new_index, sign * new_index_min)
            new[index] = sign * gs.minimum(sign * new_index, sign * new_index_max)

            symindex = k_sampling_points - 1 - index
            new_symindex = current[symindex] + step * increment[symindex - 1]
            new_symindex_min = new[index] + (k_sampling_points - 2 * index - 1) * tol
            new_symindex_max = new[symindex + 1] - tol
            new_symindex = sign * gs.minimum(
                sign * new_symindex, sign * new_symindex_max
            )
            new[symindex] = sign * gs.maximum(
                sign * new_symindex, sign * new_symindex_min
            )

        if k_sampling_points % 2 > 0:
            new_midindex = current[index + 1] + step * increment[index]
            new_midindex_min = new[index] + tol
            new_midindex_max = new[index + 2] - tol
            new_midindex = sign * gs.maximum(
                sign * new_midindex, sign * new_midindex_min
            )
            new[index + 1] = sign * gs.minimum(
                sign * new_midindex, sign * new_midindex_max
            )

        return new

    def _euler_step(self, current, increment, step):
        r"""Perform Euler step while enforcing increasing solution.

        Symmetric version of previous function.
        """
        return (
            self._euler_step_forward(current, increment, step, self.tol)
            + gs.flip(
                self._euler_step_forward(
                    gs.flip(current, axis=0),
                    gs.flip(increment, axis=0),
                    step,
                    -self.tol,
                ),
                axis=0,
            )
        ) / 2

    def _construct_reparametrization(self, vertical_norm, space_deriv_norm):
        r"""Construct path of reparametrizations.

        Construct path of reparametrizations phi(t, s) in the decomposition
        of the path of curves
        :math:`c(t, s) = c_h(t, phi(t, s))` where :math:`d/dt c_h(t, .)` is horizontal.
        This is done by solving a partial differential equation, using an Euler
        step that enforces that the solution phi(t,.) is an increasing function of
        the unit interval that preserves the end points 0 and 1, for all time t.

        Parameters
        ----------
        vertical_norm : array-like, shape=[n_times - 1, k_sampling_points]
            Pointwise norm of the vertical part of the time derivative of
            the path of curves.
        space_deriv_norm : array-like, shape=[n_times - 1, k_sampling_points]
            Pointwise norm of the space derivative of the path of curves.

        Returns
        -------
        repar : array-like, shape=[n_times, k_sampling_points]
            Path of parametrizations, such that the path of curves composed
            with the path of inverse parametrizations is a horizontal path.
        """
        n_times = vertical_norm.shape[0] + 1
        k_sampling_points = vertical_norm.shape[1]

        quotient = vertical_norm / space_deriv_norm

        repar = gs.linspace(0.0, 1.0, k_sampling_points)
        repars = [repar]
        for i in range(n_times - 1):
            repar_diff = forward_difference(repar, axis=-1)

            repar_space_deriv = gs.where(
                vertical_norm[i, 1:-1] > 0,
                repar_diff[1:],
                repar_diff[:-1],
            )
            repar_time_deriv = repar_space_deriv * quotient[i, 1:-1]
            repar = repar_i = self._euler_step(repar, repar_time_deriv, 1 / n_times)

            repars.append(repar_i)

        return gs.stack(repars)

    def _invert_reparametrize_single(self, t_space, repar, point):
        """Invert path of reparametrizations, non vectorized."""
        spline = CubicSpline(t_space, point, axis=0)
        repar_inverse = PchipInterpolator(repar, t_space)
        return gs.from_numpy(spline(repar_inverse(t_space)))

    def _invert_reparametrization(
        self,
        t_space,
        repar,
        path_of_curves,
        repar_inverse_end,
        end_spline,
    ):
        r"""Invert path of reparametrizations.

        Given a path of curves c(t, s) and a path of reparametrizations
        phi(t, s), compute:
        :math:`c(t, phi_inv(t, s))` where `phi_inv(t, .) = phi(t, .)^{-1}`

        The computation for the last time t=1 is done differently, using
        the spline function associated to the end curve and the composition
        of the inverse reparametrizations contained in rep_inverse_end:
        :math:`end_spline ° phi_inv(1, .) ° ... ° phi_inv(0, .)`.

        Parameters
        ----------
        repar : array-like, shape=[n_times, k_sampling_points]
            Path of reparametrizations.
        path_of_curves : array-like, shape=[n_times, k_sampling_points, ambient_dim]
            Path of curves.
        repar_inverse_end: list
            List of the inverses of the reparametrizations applied to
            the end curve during the optimal matching algorithm.
        end_spline : function
            Spline interpolation of the end point of the path of curves.

        Returns
        -------
        reparametrized_path : array-like, \
            shape=[n_times, k_sampling_points, ambient_dim]
            Path of curves composed with the inverse of the path of
            reparametrizations.
        """
        initial_curve = path_of_curves[0]
        reparametrized_path = [initial_curve] + [
            self._invert_reparametrize_single(
                t_space,
                repar_i,
                point,
            )
            for repar_i, point in zip(repar[1:-1], path_of_curves[1:-1])
        ]

        repar_inverse_end.append(PchipInterpolator(repar[-1], t_space))
        arg = t_space
        for repar_inverse in reversed(repar_inverse_end):
            arg = repar_inverse(arg)
        end_curve_repar = end_spline(arg)

        reparametrized_path.append(end_curve_repar)
        return gs.stack(reparametrized_path)

    def _iterate(
        self,
        times,
        t_space,
        initial_point,
        end_point,
        repar_inverse_end,
        end_spline,
    ):
        """Perform one step of the alignment algorithm."""
        ndim = self._total_space.point_ndim

        total_space_geod_fun = self._total_space.metric.geodesic(
            initial_point=initial_point, end_point=end_point
        )
        geod_points = total_space_geod_fun(times)
        geod_points_with_origin = insert_zeros(geod_points, axis=-ndim)

        time_deriv = forward_difference(geod_points, axis=-(ndim + 1))
        _, vertical_norm = self._total_space.fiber_bundle.vertical_projection(
            time_deriv, geod_points[:-1], return_norm=True
        )
        vertical_norm = insert_zeros(vertical_norm, axis=-1)

        space_deriv = centered_difference(
            geod_points_with_origin, axis=-ndim, endpoints=True
        )[:-1]

        pointwise_space_deriv_norm = self._total_space.ambient_manifold.metric.norm(
            space_deriv, geod_points_with_origin[:-1]
        )

        repar = self._construct_reparametrization(
            vertical_norm,
            pointwise_space_deriv_norm,
        )

        horizontal_path = self._invert_reparametrization(
            t_space, repar, geod_points_with_origin, repar_inverse_end, end_spline
        )

        return horizontal_path

    def _discrete_horizontal_geodesic_single(
        self, initial_point, end_point, end_spline
    ):
        """Compute discrete horizontal geodesic, non vectorized.

        Parameters
        ----------
        initial_point : array-like, shape=[k_sampling_points, ambient_dim]
            Initial discrete curve.
        end_point : array-like, shape=[k_sampling_points, ambient_dim]
            End discrete curve.
        end_spline : function
            Spline interpolation of end point.

        Returns
        -------
        horizontal_geod_points : array, shape=[n_time_grid, k - 1, ambient_dim]
            Geodesic points.
        """
        times = gs.linspace(0.0, 1.0, self.n_time_grid)

        k_sampling_points = self._total_space.k_sampling_points
        t_space = gs.linspace(0.0, 1.0, k_sampling_points)

        current_end_point = end_point
        repar_inverse_end = []
        for index in range(self.max_iter):
            horizontal_path_with_origin = self._iterate(
                times,
                t_space,
                initial_point,
                current_end_point,
                repar_inverse_end,
                end_spline,
            )
            new_end_point = horizontal_path_with_origin[-1][1:]
            l2_metric = self._total_space.discrete_curves_with_l2.metric
            gap = l2_metric.dist(new_end_point, current_end_point)
            current_end_point = new_end_point

            if gap < self.threshold:
                if self.verbose > 0:
                    logging.info(
                        f"Convergence of alignment reached after {index + 1} "
                        "iterations."
                    )
                break
        else:
            logging.warning(
                "Maximum number of iterations %d reached. The result may be inaccurate",
                self.max_iter,
            )

        if self.save_history:
            self.history = dict(
                spline=end_spline,
                repar_inverse=repar_inverse_end,
            )

        return horizontal_path_with_origin[..., 1:, :]

    def discrete_horizontal_geodesic(self, initial_point, end_point, end_spline):
        """Compute discrete horizontal geodesic.

        Parameters
        ----------
        initial_point : array-like, shape=[..., k_sampling_points, ambient_dim]
            Initial discrete curve.
        end_point : array-like, shape=[..., k_sampling_points, ambient_dim]
            End discrete curve.
        end_spline : callable or list[callable]
            Spline interpolation of end point.

        Returns
        -------
        geod_points : array, shape=[..., n_time_grid, k - 1, ambient_dim]
        """
        is_batch = check_is_batch(
            self._total_space.point_ndim,
            initial_point,
            end_point,
        )
        if not is_batch:
            return self._discrete_horizontal_geodesic_single(
                initial_point, end_point, end_spline
            )

        if initial_point.ndim != end_point.ndim:
            initial_point, end_point = gs.broadcast_arrays(initial_point, end_point)

        return gs.stack(
            [
                self._discrete_horizontal_geodesic_single(
                    initial_point_, end_point_, end_spline_
                )
                for initial_point_, end_point_, end_spline_ in zip(
                    initial_point, end_point, end_spline
                )
            ]
        )

    def align(self, point, base_point):
        """Align point to base point.

        Parameters
        ----------
        point : array-like, shape=[..., k_sampling_points - 1, ambient_dim]
            Discrete curve to align.
        base_point : array-like, shape=[..., k_sampling_points - 1, ambient_dim]
            Reference discrete curve.

        Returns
        -------
        aligned : array-like, shape=[..., k_sampling_points - 1, ambient_dim]
            Curve reparametrized in an optimal way with respect to reference curve.
        """
        if point.ndim == self._total_space.point_ndim:
            spline = self._total_space.interpolate(point)
            if base_point.ndim > self._total_space.point_ndim:
                spline = [spline] * base_point.shape[0]
        else:
            spline = [self._total_space.interpolate(point_) for point_ in point]

        return self.discrete_horizontal_geodesic(base_point, point, spline)[
            ..., -1, :, :
        ]


class DynamicProgrammingAligner(AlignerAlgorithm):
    r"""Align two curves through dynamic programming.

    Find the reparametrization gamma of end_curve that minimizes the distance
    between initial_curve and end_curve reparametrized by gamma, and output
    the corresponding distance, using a dynamic programming algorithm.

    The objective can be expressed in terms of square root velocity (SRV)
    representations: it is equivalent to finding the gamma that maximizes
    the L2 scalar product between :math:`initial_{srv}` and :math:`end_{srv}@\gamma`,
    where :math:`initial_{srv}` is the SRV representation of the initial curve
    and :math:`end_{srv}@\gamma` is the SRV representation of the end curve
    reparametrized by :math:`\gamma`, i.e

    .. math::
        end_{srv}@\gamma(t) = end_{srv}(\gamma(t))\cdot|\gamma(t)|^\frac{1}{2}

    The dynamic programming algorithm assumes that for every subinterval
    :math:`\left[\frac{i}{n},\frac{i+1}{n}\right]` of :math:`\left[0,1\right]`,
    gamma is linear.

    Parameters
    ----------
    total_space : Manifold
        Total space with reparametrizations fiber bundle structure.
    n_space_grid : int
        Number of sampling points of the unit interval.
    max_slope : int
        Maximum slope allowed for a reparametrization.

    References
    ----------
    [WAJ2007] M. Washington, S. Anuj & H. Joshi,
        "On Shape of Plane Elastic Curves", in International Journal of Computer
        Vision. 73(3):307-324, 2007.
    """

    def __init__(self, total_space, n_space_grid=100, max_slope=6):
        k_sampling_points = total_space.k_sampling_points

        super().__init__(total_space)
        self.n_space_grid = n_space_grid
        self.max_slope = max_slope

        self._srv_transform = SRVTransform(
            self._total_space.ambient_manifold,
            k_sampling_points,
        )

    def _resample_srv_function(self, srv_function):
        """Resample SRV function of a discrete curve.

        Parameters
        ----------
        srv_function : array, shape=[..., k_sampling_points - 1, ambient_dim]
            Discrete curve.

        Returns
        -------
        srv : array, shape=[..., n_space_grid - 1, ambient_dim]
            SRV function of the curve at the right size.

        Notes
        -----
        If `n_space_grid` < `k_sampling_points`, it subsamples the curve.
        """
        n_space_grid = self.n_space_grid
        index = gs.array(range(n_space_grid - 1))

        ratio = (self._total_space.k_sampling_points - 1) / (n_space_grid - 1)
        indices = gs.cast(gs.floor(index * ratio), dtype=int)

        return srv_function[..., indices, :]

    @staticmethod
    def _compute_restricted_integral(
        srv_1, srv_2, x_min_index, x_max_index, y_min_index, y_max_index
    ):
        r"""Compute the value of an integral over a subinterval.

        Compute n * the value of the integral of

        .. math::

            q_1(t)\cdot q_2(\gamma(t))\cdot|\gamma(t)|^\frac{1}{2}

        over :math:`\left[\x_min,x_max\right]` where gamma restricted to
        :math:`\left[\x_min,x_max\right]` is a linear.

        Parameters
        ----------
        srv_1 : array, shape=[n, ambient_dim]
            SRV function of the initial curve.
        srv_2 : array, shape=[n, ambient_dim]
            SRV function of the end curve.
        x_min_index : int
            Beginning of the subinterval.
        x_max_index : int
            End of the subinterval.
        y_min_index : int
            Value of gamma at x_min.
        y_max_index : int
            Value of gamma at x_max.

        Returns
        -------
        value : float
            Value of the integral described above.
        """
        delta_y_index = y_max_index - y_min_index
        delta_x_index = x_max_index - x_min_index

        gamma_slope = delta_y_index / delta_x_index

        x_bound_indices = list(range(x_min_index, x_max_index + 1))
        y_bounds_indices = [
            (y_index - y_min_index) / gamma_slope + x_min_index
            for y_index in range(y_min_index, y_max_index + 1)
        ]

        lower_bound = x_min_index
        x_index, y_index = 0, 0

        value = 0.0
        while x_index < delta_x_index and y_index < delta_y_index:
            upper_bound = min(
                x_bound_indices[x_index + 1], y_bounds_indices[y_index + 1]
            )

            # NB: normalization is done outside
            length = upper_bound - lower_bound
            value += length * gs.dot(
                srv_1[x_min_index + x_index], srv_2[y_min_index + y_index]
            )

            if (
                gs.abs(x_bound_indices[x_index + 1] - y_bounds_indices[y_index + 1])
                < gs.atol
            ):
                x_index += 1
                y_index += 1
            elif x_bound_indices[x_index + 1] < y_bounds_indices[y_index + 1]:
                x_index += 1
            else:
                y_index += 1
            lower_bound = upper_bound

        return gamma_slope**0.5 * value

    def _reparametrize(self, point_with_origin, gamma):
        """Reparametrize curve by gamma.

        Parameters
        ----------
        point_with_origin : array, shape=[k_sampling_points, ambient_dim]
            Discrete curve.
        gamma : array, shape=[n_subinterval]
            Parametrization of a curve.

        Returns
        -------
        new_point : array , shape=[k_sampling_points - 1, ambient_dim]
            Curve reparametrized by gamma.
        """
        n_space_grid = self.n_space_grid
        k_sampling_points = self._total_space.k_sampling_points

        x_uniform = gs.linspace(0.0, 1.0, k_sampling_points)
        point_interpolator = UniformUnitIntervalLinearInterpolator(
            point_with_origin, point_ndim=1
        )

        x_space = gs.array([gamma_elem[1] for gamma_elem in gamma]) / (n_space_grid - 1)
        y_space = gs.array([gamma_elem[0] for gamma_elem in gamma]) / (n_space_grid - 1)

        repar_inverse = LinearInterpolator1D(y_space, x_space, point_ndim=0)

        return point_interpolator(repar_inverse(x_uniform[1:]))

    def _compute_squared_dist(self, initial_srv, end_srv, grid_last):
        """Compute squared distance using algorithmic information."""
        n_space_grid = self.n_space_grid

        norm_squared_initial_srv = self._compute_restricted_integral(
            initial_srv, initial_srv, 0, n_space_grid - 1, 0, n_space_grid - 1
        ) / (n_space_grid - 1)
        norm_squared_end_srv = self._compute_restricted_integral(
            end_srv, end_srv, 0, n_space_grid - 1, 0, n_space_grid - 1
        ) / (n_space_grid - 1)

        maximum_scalar_product = grid_last / (n_space_grid - 1)
        return (
            norm_squared_initial_srv + norm_squared_end_srv - 2 * maximum_scalar_product
        )

    def _align_single(self, point, base_point, return_sdist=False):
        r"""Align point to base point, non vectorized.

        Parameters
        ----------
        point : array-like, shape=[k_sampling_points - 1, ambient_dim]
            Discrete curve to align.
        base_point : array-like, shape=[k_sampling_points - 1, ambient_dim]
            Reference discrete curve.
        return_sdist : bool
            If True, also returns squared distance.

        Returns
        -------
        aligned : array, shape=[k_sampling_points - 1, ambient_dim]
            Curve reparametrized in an optimal way with respect to reference curve.
        squared_dist : float
            Quotient distance between point and base point.
            If return_sdist is True.
        """
        max_slope = self.max_slope
        n_space_grid = self.n_space_grid

        initial_srv = self._srv_transform(base_point)
        end_srv = self._srv_transform(point)

        resampled_initial_srv = self._resample_srv_function(initial_srv)
        resampled_end_srv = self._resample_srv_function(end_srv)

        grid = {(0, 0): 0.0}
        gamma = {(0, 0): [(0, 0)]}

        for x_max_index in range(1, n_space_grid):
            max_y_max_index = max(n_space_grid, x_max_index + 1)
            for y_max_index in range(1, max_y_max_index):
                # NB: solves minimization part
                min_x_min_index = max(0, x_max_index - max_slope)
                min_y_min_index = max(0, y_max_index - max_slope)
                for x_min_index in range(min_x_min_index, x_max_index):
                    for y_min_index in range(min_y_min_index, y_max_index):
                        if (x_min_index, y_min_index) not in grid:
                            # jump borders
                            continue
                        elif (
                            gs.abs(
                                (y_max_index - y_min_index)
                                / (x_max_index - x_min_index)
                                - 1
                            )
                            < gs.atol
                            and x_max_index - x_min_index != 1
                            and y_max_index - y_max_index != 1
                        ):
                            # jump slope == 1 except last
                            continue

                        new_value = grid[
                            x_min_index, y_min_index
                        ] + self._compute_restricted_integral(
                            resampled_initial_srv,
                            resampled_end_srv,
                            x_min_index,
                            x_max_index,
                            y_min_index,
                            y_max_index,
                        )
                        if (x_max_index, y_max_index) not in grid or grid[
                            (x_max_index, y_max_index)
                        ] < new_value:
                            grid[x_max_index, y_max_index] = new_value
                            new_gamma = gamma[(x_min_index, y_min_index)].copy()
                            new_gamma.append((x_max_index, y_max_index))
                            gamma[(x_max_index, y_max_index)] = new_gamma

        last_index = n_space_grid - 1
        point_with_origin = insert_zeros(point, axis=-2)
        point_reparametrized = self._reparametrize(
            point_with_origin, gamma[last_index, last_index]
        )

        if not return_sdist:
            return point_reparametrized

        return point_reparametrized, self._compute_squared_dist(
            resampled_initial_srv,
            resampled_end_srv,
            grid[(last_index, last_index)],
        )

    def align(self, point, base_point, return_sdist=False):
        """Align point to base point.

        Parameters
        ----------
        point : array-like, shape=[..., k_sampling_points - 1, ambient_dim]
            Discrete curve to align.
        base_point : array-like, shape=[..., k_sampling_points - 1, ambient_dim]
            Reference discrete curve.
        return_sdist : bool
            If True, also returns squared distance.

        Returns
        -------
        aligned : array-like, shape=[..., k_sampling_points - 1, ambient_dim]
            Curve reparametrized in an optimal way with respect to reference curve.
        squared_dist : array, shape=[...,]
            Quotient distance between point and base point.
            If return_sdist is True.
        """
        is_batch = check_is_batch(
            self._total_space.point_ndim,
            point,
            base_point,
        )
        if not is_batch:
            return self._align_single(point, base_point, return_sdist=return_sdist)

        if point.ndim != base_point.ndim:
            point, base_point = gs.broadcast_arrays(point, base_point)

        out = [
            self._align_single(point_, base_point_, return_sdist=return_sdist)
            for point_, base_point_ in zip(point, base_point)
        ]
        if not return_sdist:
            return gs.stack(out)

        aligned = gs.stack([out_[0] for out_ in out])
        sdists = gs.stack([out_[1] for out_ in out])
        return aligned, sdists


class ReparametrizationBundle(FiberBundle):
    """Principal bundle of curves modulo reparametrizations with an elastic metric.

    The space of parameterized curves is the total space of a principal bundle
    where the group action is given by reparametrization and the base space is
    the shape space of curves modulo reparametrization, i.e.unparametrized
    curves. In the discrete case, reparametrization corresponds to resampling.

    Each tangent vector to the space of parameterized curves can be split into a
    vertical part (tangent to the fibers of the principal bundle) and a
    horizontal part (orthogonal to the vertical part with respect to the SRV
    metric). The geodesic between the shapes of two curves is computed by
    aligning (i.e. reparametrizing) one of the two curves with respect to the
    other, and computing the geodesic between the aligned curves. This geodesic
    will be horizontal, and will project to a geodesic on the shape space.

    Two different aligners are available:
    - IterativeHorizontalGeodesicAligner (default)
    - DynamicProgrammingAligner.

    Parameters
    ----------
    total_space : DiscreteCurvesStartingAtOrigin
        Space of discrete curves starting at the origin
    """

    def __init__(self, total_space, aligner=None):
        if aligner is None:
            aligner = IterativeHorizontalGeodesicAligner(total_space)

        super().__init__(
            total_space=total_space,
            aligner=aligner,
        )

    def vertical_projection(self, tangent_vec, base_point, return_norm=False):
        """Compute vertical part of tangent vector at base point.

        Parameters
        ----------
        tangent_vec : array-like, shape=[..., k_sampling_points - 1, ambient_dim]
            Tangent vector to decompose into horizontal and vertical parts.
        base_point : array-like, shape=[..., k_sampling_points - 1, ambient_dim]
            Discrete curve, base point of tangent_vec in the manifold of curves.
        return_norm : boolean,
            If True, the method returns the pointwise norm of the vertical
            part of tangent_vec.

        Returns
        -------
        tangent_vec_ver : array-like, shape=[..., k_sampling_points - 1, ambient_dim]
            Vertical part of tangent_vec.
        vertical_norm: array-like, shape=[..., n_points]
            Pointwise norm of the vertical part of tangent_vec.
            Only returned when return_norm is True.
        """
        ambient_manifold = self._total_space.ambient_manifold

        a_param, b_param = self._total_space.metric.a, self._total_space.metric.b
        squotient = (a_param / b_param) ** 2

        ndim = self._total_space.point_ndim
        tangent_vec_with_zeros = insert_zeros(tangent_vec, axis=-ndim)
        base_point_with_origin = insert_zeros(base_point, axis=-ndim)

        position = base_point_with_origin[..., 1:-1, :]
        delta = 1.0
        d_pos = centered_difference(base_point_with_origin, delta=delta, axis=-ndim)
        d_vec = centered_difference(tangent_vec_with_zeros, delta=delta, axis=-ndim)
        d2_pos = second_centered_difference(
            base_point_with_origin, delta=delta, axis=-ndim
        )
        d2_vec = second_centered_difference(
            tangent_vec_with_zeros, delta=delta, axis=-ndim
        )

        pointwise_snorm = ambient_manifold.metric.squared_norm(
            d_pos,
            position,
        )
        pointwise_norm = gs.sqrt(pointwise_snorm)
        pointwise_snorm2 = ambient_manifold.metric.squared_norm(
            d2_pos,
            position,
        )
        pointwise_inner_product_pos = ambient_manifold.metric.inner_product(
            d2_pos, d_pos, position
        )

        vec_a = pointwise_snorm - 1 / 2 * pointwise_inner_product_pos
        vec_b = -2 * pointwise_snorm - squotient * (
            pointwise_snorm2 - pointwise_inner_product_pos**2 / pointwise_snorm
        )
        vec_c = pointwise_snorm + 1 / 2 * pointwise_inner_product_pos
        vec_d = pointwise_norm * (
            ambient_manifold.metric.inner_product(d2_vec, d_pos, position)
            - (squotient - 1)
            * ambient_manifold.metric.inner_product(d_vec, d2_pos, position)
            + (squotient - 2)
            * ambient_manifold.metric.inner_product(d2_pos, d_pos, position)
            * ambient_manifold.metric.inner_product(d_vec, d_pos, position)
            / pointwise_snorm
        )

        linear_system = (
            from_vector_to_diagonal_matrix(vec_a[..., :-1], 1)
            + from_vector_to_diagonal_matrix(vec_b, 0)
            + from_vector_to_diagonal_matrix(vec_c[..., 1:], -1)
        )
        if linear_system.ndim == 2 and tangent_vec.ndim > 2:
            linear_system = gs.broadcast_to(
                linear_system, vec_d.shape[:-1] + linear_system.shape
            )

        vertical_norm = gs.linalg.solve(linear_system, vec_d)

        unit_speed = gs.einsum(
            "...ij,...i->...ij",
            d_pos,
            1 / pointwise_norm,
        )
        tangent_vec_ver = gs.einsum("...ij,...i->...ij", unit_speed, vertical_norm)

        tangent_vec_ver = insert_zeros(tangent_vec_ver, axis=-ndim, end=True)
        if return_norm:
            vertical_norm = insert_zeros(vertical_norm, axis=-1, end=True)
            return tangent_vec_ver, vertical_norm

        return tangent_vec_ver


class RotationBundle(FiberBundle):
    """Principal bundle of curves modulo rotations with an elastic metric.

    This is the fiber bundle where the total space is the space of parameterized
    curves equipped with an elastic metric, the action is given by rotations, and
    the base space is the shape space of curves modulo rotations.

    Parameters
    ----------
    total_space : DiscreteCurvesStartingAtOrigin
        Space of discrete curves starting at the origin
    """

    def _rotate(self, point, rotation):
        """Rotate discrete curve starting at origin."""
        return Matrices.transpose(gs.matmul(rotation, Matrices.transpose(point)))

    def align(self, point, base_point, return_rotation=False):
        """Align point to base point.

        Find optimal rotation of curve with respect to base curve.

        Parameters
        ----------
        point : array-like, shape=[..., k_sampling_points - 1, ambient_dim]
            Discrete curve to align.
        base_point : array-like, shape=[..., k_sampling_points - 1, ambient_dim]
            Reference discrete curve.
        return_rotation : boolean
            If true, returns the optimal rotation used for the alignment.
            Optional, default : False.

        Returns
        -------
        aligned : array-like, shape=[..., k_sampling_points - 1, ambient_dim
            Curve optimally rotated with respect to reference curve.
        """
        transform = self._total_space.metric.diffeo
        initial_srv = transform(base_point)
        end_srv = transform(point)

        mat = gs.matmul(Matrices.transpose(initial_srv), end_srv)
        u_svd, _, vt_svd = gs.linalg.svd(mat)
        sign = gs.linalg.det(gs.matmul(u_svd, vt_svd))
        vt_svd[..., -1, :] = gs.einsum("...,...j->...j", sign, vt_svd[..., -1, :])
        rotation = gs.matmul(u_svd, vt_svd)

        point_aligned = self._rotate(point, rotation)
        if return_rotation:
            return point_aligned, rotation

        return point_aligned


register_quotient(
    Space=DiscreteCurvesStartingAtOrigin,
    Metric=SRVMetric,
    GroupAction="rotations",
    FiberBundle=RotationBundle,
    QuotientMetric=QuotientMetric,
)
register_quotient(
    Space=DiscreteCurvesStartingAtOrigin,
    Metric=SRVMetric,
    GroupAction="reparametrizations",
    FiberBundle=ReparametrizationBundle,
    QuotientMetric=QuotientMetric,
)



import pytest

import geomstats.backend as gs
from geomstats.test_cases.geometry.complex_riemannian_metric import (
    ComplexRiemannianMetricTestCase,
)


class SiegelMetricTestCase(ComplexRiemannianMetricTestCase):
    def test_tangent_vec_from_base_point_to_zero(
        self, tangent_vec, base_point, expected, atol
    ):
        tangent_vec_at_zero = self.space.metric.tangent_vec_from_base_point_to_zero(
            tangent_vec, base_point
        )
        self.assertAllClose(tangent_vec_at_zero, expected, atol=atol)

    @pytest.mark.random
    def test_tangent_vec_from_base_point_to_zero_is_tangent(self, n_points, atol):
        base_point = self.data_generator.random_point(n_points)
        tangent_vec = self.data_generator.random_tangent_vec(base_point)

        tangent_vec_at_zero = self.space.metric.tangent_vec_from_base_point_to_zero(
            tangent_vec, base_point
        )
        zero = gs.zeros(self.space.shape)
        is_tangent = self.space.is_tangent(tangent_vec_at_zero, zero, atol=atol)

        expected = gs.ones(n_points, dtype=bool)
        self.assertAllEqual(is_tangent, expected)

    # def test_sectional_curvature_at_zero(
    #     self, tangent_vec_a, tangent_vec_b, base_point, expected, atol
    # ):
    #     res = self.space.metric.sectional_curvature(
    #         tangent_vec_a, tangent_vec_b, base_point
    #     )
    #     self.assertAllClose(res, expected, atol=atol)



import math
import random

import pytest

import geomstats.backend as gs
from geomstats._mesh import Surface
from geomstats.test.test_case import TestCase
from geomstats.test_cases.geometry.mixins import DistTestCaseMixins

if gs.__name__.endswith("pytorch"):
    from pykeops.torch import Genred
else:
    from pykeops.numpy import Genred


class VarifoldRandomDataGenerator:
    def __init__(self, point, amplitude=2.0, max_vertices_removal=2):
        self.point = point
        self.max_vertices_removal = max_vertices_removal
        self.amplitude = amplitude

    def random_point(self, n_points=1):
        if n_points != 1:
            raise ValueError("Can only generate one point at a time.")

        n_remove = random.randint(0, self.max_vertices_removal)
        new_vertices = self.point.vertices
        new_faces = self.point.faces
        for _ in range(n_remove):
            new_vertices, new_faces = self._remove_last_vertex(new_vertices, new_faces)

        new_vertices += self._get_deformation(new_vertices.shape[0])
        return Surface(new_vertices, new_faces)

    def _remove_last_vertex(self, vertices, faces):
        index = vertices.shape[0] - 1
        indices = gs.array([index not in face for face in faces])
        return vertices[:-1], faces[indices]

    def _get_deformation(self, n_vertices):
        rand_shape = (n_vertices, 3)
        return gs.reshape(
            gs.random.rand(math.prod(rand_shape)),
            rand_shape,
        )


class KernelTestCase(TestCase):
    def test_against_other(self, point_a, point_b, atol):
        res = self.kernel(point_a, point_b)
        res_ = self.kernel_(point_a, point_b)
        self.assertAllClose(res, res_, atol=atol)

    @pytest.mark.random
    def test_against_other_random(self, n_points, atol):
        point_a = self.data_generator.random_point(n_points)
        point_b = self.data_generator.random_point(n_points)

        self.test_against_other(point_a, point_b, atol)


class VarifoldMetricTestCase(DistTestCaseMixins, TestCase):
    pass


def GaussianKernel_(sigma):
    expr = Genred(
        "Exp(-SqDist(x,y)*a)",
        [
            "a=Pm(1)",
            "x=Vi(3)",
            "y=Vj(3)",
        ],
        reduction_op="Sum",
        axis=1,
    )
    a_param = 1 / gs.array([sigma]) ** 2

    def kernel_eval(point_a, point_b):
        print(type(a_param), type(point_a), type(point_b))
        return expr(a_param, point_a, point_b)

    return kernel_eval


def CauchyKernel_(sigma):
    expr = Genred(
        "IntCst(1)/(IntCst(1)+SqDist(x,y)*a)",
        [
            "a=Pm(1)",
            "x=Vi(3)",
            "y=Vj(3)",
        ],
        reduction_op="Sum",
        axis=1,
    )
    a_param = 1 / gs.array([sigma]) ** 2

    def kernel_eval(point_a, point_b):
        return expr(a_param, point_a, point_b)

    return kernel_eval


def LinearKernel_():
    expr = Genred(
        "(u|v)",
        [
            "u=Vi(3)",
            "v=Vj(3)",
        ],
        reduction_op="Sum",
        axis=1,
    )

    def kernel_eval(point_a, point_b):
        return expr(point_a, point_b)

    return kernel_eval


def BinetKernel_():
    expr = Genred(
        "Square((u|v))",
        [
            "u=Vi(3)",
            "v=Vj(3)",
        ],
        reduction_op="Sum",
        axis=1,
    )

    def kernel_eval(point_a, point_b):
        return expr(point_a, point_b)

    return kernel_eval


def OrientedGaussianKernel_(sigma=1.0):
    expr = Genred(
        "Exp(IntCst(2)*b*((u|v)-IntCst(1)))",
        [
            "b=Pm(1)",
            "u=Vi(3)",
            "v=Vj(3)",
        ],
        reduction_op="Sum",
        axis=1,
    )

    a_param = 1 / gs.array([sigma]) ** 2

    def kernel_eval(point_a, point_b):
        return expr(a_param, point_a, point_b)

    return kernel_eval


def UnorientedGaussianKernel_(sigma=1.0):
    expr = Genred(
        "Exp(IntCst(2)*b*(Square((u|v))-IntCst(1)))",
        [
            "b=Pm(1)",
            "u=Vi(3)",
            "v=Vj(3)",
        ],
        reduction_op="Sum",
        axis=1,
    )

    a_param = 1 / gs.array([sigma]) ** 2

    def kernel_eval(point_a, point_b):
        return expr(a_param, point_a, point_b)

    return kernel_eval



import pytest

from geomstats.test.vectorization import generate_vectorization_data
from geomstats.test_cases.geometry.base import MatrixVectorSpaceTestCase


class MatrixLieAlgebraTestCase(MatrixVectorSpaceTestCase):
    def test_baker_campbell_hausdorff(
        self, matrix_a, matrix_b, expected, atol, order=2
    ):
        res = self.space.baker_campbell_hausdorff(matrix_a, matrix_b, order=order)
        self.assertAllClose(res, expected, atol=atol)

    @pytest.mark.vec
    def test_baker_campbell_hausdorff_vec(self, n_reps, atol, order=2):
        matrix_a, matrix_b = self.data_generator.random_point(2)
        expected = self.space.baker_campbell_hausdorff(matrix_a, matrix_b, order=order)

        vec_data = generate_vectorization_data(
            data=[
                dict(
                    matrix_a=matrix_a,
                    matrix_b=matrix_b,
                    expected=expected,
                    atol=atol,
                    order=order,
                )
            ],
            arg_names=["matrix_a", "matrix_b"],
            expected_name="expected",
            n_reps=n_reps,
        )
        self._test_vectorization(vec_data)



import functools

import torch as _torch
from torch import complex64, complex128, float32, float64

from geomstats._backend import _backend_config as _config
from geomstats._backend._dtype_utils import (
    _MAP_FLOAT_TO_COMPLEX,
    _modify_func_default_dtype,
    _pre_add_default_dtype_by_casting,
    _pre_allow_complex_dtype,
    _pre_cast_out_to_input_dtype,
    _update_default_dtypes,
    get_default_cdtype,
    get_default_dtype,
)

from ._common import cast

MAP_DTYPE = {
    "float32": float32,
    "float64": float64,
    "complex64": complex64,
    "complex128": complex128,
}

_COMPLEX_DTYPES = (complex64, complex128)


def is_floating(x):
    return x.dtype.is_floating_point


def is_complex(x):
    return x.dtype.is_complex


def is_bool(x):
    return x.dtype is _torch.bool


def as_dtype(value):
    """Transform string representing dtype in dtype."""
    return MAP_DTYPE[value]


def _dtype_as_str(dtype):
    return str(dtype).split(".")[-1]


def set_default_dtype(value):
    """Set backend default dtype.

    Parameters
    ----------
    value : str
        Possible values are "float32" as "float64".
    """
    _config.DEFAULT_DTYPE = as_dtype(value)
    _config.DEFAULT_COMPLEX_DTYPE = as_dtype(_MAP_FLOAT_TO_COMPLEX.get(value))
    _torch.set_default_dtype(_config.DEFAULT_DTYPE)

    _update_default_dtypes()

    return _config.DEFAULT_DTYPE


_add_default_dtype_by_casting = _pre_add_default_dtype_by_casting(cast)
_cast_out_to_input_dtype = _pre_cast_out_to_input_dtype(
    cast, is_floating, is_complex, as_dtype, _dtype_as_str
)
_allow_complex_dtype = _pre_allow_complex_dtype(cast, _COMPLEX_DTYPES)


def _preserve_input_dtype(target=None):
    """Ensure input dtype is preserved.

    How it works?
    -------------
    Only acts on input. Assumes dtype is kwarg and function accepts dtype.
    Passes dtype as input dtype.

    Use together with `_add_default_dtype_by_casting`.
    """

    def _decorator(func):
        @functools.wraps(func)
        def _wrapped(x, *args, dtype=None, **kwargs):
            if dtype is None:
                dtype = x.dtype

            return func(x, *args, dtype=dtype, **kwargs)

        return _wrapped

    if target is None:
        return _decorator

    return _decorator(target)


def _box_unary_scalar(target=None):
    """Update dtype if input is float in unary operations.

    How it works?
    -------------
    Promotes input to tensor if not the case.
    """

    def _decorator(func):
        @functools.wraps(func)
        def _wrapped(x, *args, **kwargs):
            if not _torch.is_tensor(x):
                x = _torch.tensor(x)
            return func(x, *args, **kwargs)

        return _wrapped

    if target is None:
        return _decorator

    return _decorator(target)


def _box_binary_scalar(target=None, box_x1=True, box_x2=True):
    """Update dtype if input is float in binary operations.

    How it works?
    -------------
    Promotes inputs to tensor if not the case.
    """

    def _decorator(func):
        @functools.wraps(func)
        def _wrapped(x1, x2, *args, **kwargs):
            if box_x1 and not _torch.is_tensor(x1):
                x1 = _torch.tensor(x1)
            if box_x2 and not _torch.is_tensor(x2):
                x2 = _torch.tensor(x2)

            return func(x1, x2, *args, **kwargs)

        return _wrapped

    if target is None:
        return _decorator

    return _decorator(target)



import pytest

import geomstats.backend as gs
from geomstats.test.random import RandomDataGenerator, get_random_times
from geomstats.test.test_case import TestCase
from geomstats.test.vectorization import generate_vectorization_data
from geomstats.test_cases.geometry.stratified.point_set import (
    PointSetMetricWithArrayTestCase,
)


class AlignerAlgorithmTestCase(TestCase):
    # TODO: add align against itself test
    def setup_method(self):
        if not hasattr(self, "data_generator"):
            self.data_generator = RandomDataGenerator(self.total_space)

    def test_align(self, point, base_point, expected, atol):
        res = self.aligner.align(point, base_point)
        self.assertAllClose(res, expected, atol=atol)

    @pytest.mark.vec
    def test_align_vec(self, n_reps, atol):
        # TODO: extend vec_test generation to accept this case?
        point = self.data_generator.random_point()
        base_point = self.data_generator.random_point()

        expected = self.aligner.align(point, base_point)

        vec_data = generate_vectorization_data(
            data=[
                dict(point=point, base_point=base_point, expected=expected, atol=atol)
            ],
            arg_names=["point", "base_point"],
            expected_name="expected",
            n_reps=n_reps,
        )
        self._test_vectorization(vec_data)


class AlignerAlgorithmCmpTestCase(TestCase):
    def setup_method(self):
        if not hasattr(self, "data_generator"):
            self.data_generator = RandomDataGenerator(self.total_space)

    @pytest.mark.random
    def test_align(self, n_points, atol):
        point = self.data_generator.random_point(n_points)
        base_point = self.data_generator.random_point(n_points)

        aligned = self.aligner.align(point, base_point)
        other_aligned = self.other_aligner.align(point, base_point)

        self.assertAllClose(aligned, other_aligned, atol=atol)


class QuotientMetricWithArrayTestCase(PointSetMetricWithArrayTestCase):
    """Quotient metric test cases.

    Applies to discrete group actions.
    """

    @pytest.mark.random
    def test_geodesic_boundary_points(self, n_points, atol):
        initial_point = self.data_generator.random_point(n_points)
        end_point = self.data_generator.random_point(n_points)

        time = gs.array([0.0, 1.0])

        geod_func = self.space.metric.geodesic(initial_point, end_point=end_point)

        res = geod_func(time)
        aligned_end_point = self.total_space.aligner.align(end_point, initial_point)
        expected = gs.stack(
            [initial_point, aligned_end_point], axis=-(self.space.point_ndim + 1)
        )
        self.assertAllClose(res, expected, atol=atol)

    @pytest.mark.random
    def test_geodesic_bvp_reverse(self, n_points, n_times, atol):
        initial_point = self.data_generator.random_point(n_points)
        end_point = self.data_generator.random_point(n_points)

        time = get_random_times(n_times)

        geod_func = self.space.metric.geodesic(initial_point, end_point=end_point)
        geod_func_reverse = self.space.metric.geodesic(
            end_point, end_point=initial_point
        )

        res = geod_func(time)
        res_ = geod_func_reverse(1.0 - time)
        if n_points > 1:
            aligned_res_ = gs.stack(
                [
                    self.total_space.aligner.align(point_res_, initial_point_)
                    for point_res_, initial_point_ in zip(res_, initial_point)
                ]
            )
        else:
            aligned_res_ = self.total_space.aligner.align(res_, initial_point)

        self.assertAllClose(res, aligned_res_, atol=atol)



"""Visualization for Geometric Statistics."""

import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D  # NOQA

import geomstats.backend as gs


class Ellipses:
    """Class used to plot points on the manifold SPD(2).

    Elements S of the manifold of 2D Symmetric Positive Definite matrices
    can be conveniently represented by ellipses.

    We write :math:`S = O D O^T` with :math:`O` an orthogonal matrix (rotation)
    and :math:`D` a diagonal matrix. The positive eigenvalues, i.e. the elements of
    :math:`D`, are the inverse of the length of the major and minor axes of
    the ellipse. The rotation matrix :math:`O` determines the orientation of the
    2D ellipse in the 2D plane.

    Parameters
    ----------
    k_sampling_points : int
        Number of points to sample on the discretized ellipses.
    """

    def __init__(self, k_sampling_points=100):
        self.k_sampling_points = k_sampling_points

    @staticmethod
    def set_ax(ax=None):
        """Set the axis for the Figure.

        Parameters
        ----------
        ax : Axis
            Axis of the figure.

        Returns
        -------
        ax : Axis
            Axis of the figure.
        """
        if ax is None:
            ax = plt.subplot()
        plt.setp(ax, xlabel="X", ylabel="Y")
        return ax

    def draw_points(self, points=None, ax=None, **plot_kwargs):
        """Draw the ellipses.

        Parameters
        ----------
        ax : Axis
            Axis of the figure.
        points : array-like, shape=[..., 2, 2]
            Points on the SPD manifold of 2D symmetric
            positive definite matrices.
            Optional, default: None.
        plot_kwargs : dict
            Dictionary of arguments related to plotting.
        """
        if ax is None:
            ax = self.set_ax()
        if points.ndim == 2:
            points = [points]
        for point in points:
            x_coords, y_coords = self.compute_coordinates(point)
            ax.plot(x_coords, y_coords, **plot_kwargs)

    def compute_coordinates(self, point):
        """Compute the ellipse coordinates of a 2D SPD matrix.

        Parameters
        ----------
        point : array-like, shape=[2, 2]
            SPD matrix.

        Returns
        -------
        x_coords : array-like, shape=[k_sampling_points,]
            x_coords coordinates of the sampling points on the discretized ellipse.
        Y: array-like, shape = [k_sampling_points,]
            y coordinates of the sampling points on the discretized ellipse.
        """
        eigvalues, eigvectors = gs.linalg.eigh(point)
        eigvalues = gs.where(eigvalues < gs.atol, gs.atol, eigvalues)

        [eigvalue1, eigvalue2] = eigvalues

        rot_sin = eigvectors[1, 0]
        rot_cos = eigvectors[0, 0]
        thetas = gs.linspace(0.0, 2 * gs.pi, self.k_sampling_points + 1)

        x_coords = eigvalue1 * gs.cos(thetas) * rot_cos
        x_coords -= rot_sin * eigvalue2 * gs.sin(thetas)
        y_coords = eigvalue1 * gs.cos(thetas) * rot_sin
        y_coords += rot_cos * eigvalue2 * gs.sin(thetas)
        return x_coords, y_coords



"""The manifold of positive reals.

This manifold is a particular case in dimension 1
of the SPD matrices endowed with the SPD affine-invariant metric
with a power affine coefficient equal to one.
This manifold is used as part of a product manifold involving
complex Poincare disks to classify complex stationary centered
Gaussian autoregressive time series in [Cabanes2022]_, [Cabanes_CESAR_2019]_,
[Cabanes_GSI_2019]_ and [Cabanes_RADAR_2019]_.
This product manifold is called ProductPositiveRealsAndComplexPoincareDisks
in geomstats. It is a product manifold of complex type, therefore the manifold
PositiveReals is compatible with complex input points and vectors even if the
input values should be reals.


Lead author: Yann Cabanes.

References
----------
.. [Cabanes2022] Yann Cabanes. Multidimensional complex stationary
    centered Gaussian autoregressive time series machine learning
    in Poincaré and Siegel disks: application for audio and radar
    clutter classification, PhD thesis, 2022
.. [Cabanes_CESAR_2019] Yann Cabanes, Frédéric Barbaresco, Marc Arnaudon,
    Jérémie Bigot. Unsupervised Machine Learning for Pathological Radar
    Clutter Clustering: the P-Mean-Shift Algorithm, C&ESAR 2019, Nov 2019,
    Rennes, France. hal-02875430
    https://hal.archives-ouvertes.fr/hal-02875430/document
.. [Cabanes_GSI_2019] Yann Cabanes, Frédéric Barbaresco, Marc Arnaudon,
    Jérémie Bigot. Toeplitz Hermitian Positive Definite Matrix Machine Learning
    based on Fisher Metric. Geometric Science of Information, 2019,
    10.1007/978-3-030-26980-7_27. hal-02875403
    https://hal.archives-ouvertes.fr/hal-02875403/document
.. [Cabanes_RADAR_2019] Yann Cabanes, Frédéric Barbaresco, Marc Arnaudon,
    Jérémie Bigot. Non-Supervised High Resolution Doppler Machine Learning
    for Pathological Radar Clutter. RADAR 2019, Sep 2020, Toulon, France.
    10.1109/RADAR41533.2019.171295. hal-02875415
    https://hal.archives-ouvertes.fr/hal-02875415/document
.. [JV2016] B. Jeuris and R. Vandebril. The Kahler mean of Block-Toeplitz
    matrices with Toeplitz structured blocks, 2016.
    https://epubs.siam.org/doi/pdf/10.1137/15M102112X
"""

import geomstats.backend as gs
from geomstats.geometry.base import VectorSpaceOpenSet
from geomstats.geometry.euclidean import Euclidean
from geomstats.geometry.riemannian_metric import RiemannianMetric


class PositiveReals(VectorSpaceOpenSet):
    """Class for the manifold of positive reals.

    The real positive axis endowed with the Information geometry metric.
    """

    def __init__(self, equip=True):
        super().__init__(
            dim=1,
            embedding_space=Euclidean(1),
            equip=equip,
        )

    @staticmethod
    def default_metric():
        """Metric to equip the space with if equip is True."""
        return PositiveRealsMetric

    @staticmethod
    def belongs(point, atol=gs.atol):
        """Check if a point is a positive real.

        Evaluate if a point belongs to the positive real axis,
        i.e. evaluate if its norm is lower than one.

        Parameters
        ----------
        point : array-like, shape=[..., 1]
            Points to be checked.
        atol : float
            Tolerance.
            Optional, default: backend atol.

        Returns
        -------
        belongs : array-like, shape=[...]
            Boolean denoting if point is a positive real.
        """
        if point.shape[-1] != 1:
            return gs.zeros(point.shape[:-1], dtype=bool)

        is_real = gs.all(gs.abs(gs.imag(point)) < atol, axis=(-1,))
        is_positive = gs.all(gs.real(point) > 0, axis=(-1,))

        res = gs.logical_and(is_positive, is_real)
        if not gs.is_array(res):
            return gs.array(res)

        return res

    @staticmethod
    def projection(point, atol=gs.atol):
        """Project a point on the positive reals.

        Parameters
        ----------
        point : array-like, shape=[..., 1]
            Point in ambient manifold.
        atol : float
            Tolerance.
            Optional, default: backend atol.

        Returns
        -------
        projected : array-like, shape=[..., 1]
            Projected point.
        """
        return gs.where(point <= 0, atol, point)

    @staticmethod
    def random_point(n_samples=1, bound=1.0, atol=gs.atol):
        """Sample in the positive reals.

        Parameters
        ----------
        n_samples : int
            Number of samples.
            Optional, default: 1.
        bound : float
            Bound of the interval in which to sample in the tangent space.
            Optional, default: 1.
        atol : float
            Tolerance.
            Optional, default: backend atol.

        Returns
        -------
        samples : array-like, shape=[..., 1]
            Points sampled in the positive reals.
        """
        size = (n_samples, 1) if n_samples != 1 else (1,)
        return (bound - atol) * gs.random.rand(*size) + atol


class PositiveRealsMetric(RiemannianMetric):
    """Class for the positive reals metric.

    It is a particular case in dimension 1
    of the SPD affine-invariant metric
    with a power affine coefficient equal to one.
    """

    def metric_matrix(self, base_point):
        """Compute the inner product matrix at base point.

        Parameters
        ----------
        base_point : array-like, shape=[..., 1]
            Base point.

        Returns
        -------
        inner_prod_mat : array-like, shape=[..., 1, 1]
            Inner product matrix.
        """
        inner_prod_mat = 1 / base_point**2
        return gs.expand_dims(inner_prod_mat, axis=-1)

    @staticmethod
    def exp(tangent_vec, base_point):
        """Compute the positive reals exponential map.

        Parameters
        ----------
        tangent_vec : array-like, shape=[..., 1]
            Tangent vector at base point.
        base_point : array-like, shape=[..., 1]
            Base point.

        Returns
        -------
        exp : array-like, shape=[..., 1]
            Riemannian exponential.
        """
        return base_point * gs.exp(tangent_vec / base_point)

    @staticmethod
    def log(point, base_point):
        """Compute the positive reals logarithm map.

        Compute the Riemannian logarithm at point base_point,
        of point wrt the positive reals metric.
        This gives a tangent vector at point base_point.

        Parameters
        ----------
        point : array-like, shape=[..., 1]
            Point.
        base_point : array-like, shape=[..., 1]
            Base point.

        Returns
        -------
        log : array-like, shape=[..., 1]
            Riemannian logarithm.
        """
        return base_point * gs.log(point / base_point)

    def squared_dist(self, point_a, point_b):
        """Compute the positive reals squared distance.

        Compute the Riemannian squared distance between point_a and point_b.

        Parameters
        ----------
        point_a : array-like, shape=[..., 1]
            Point.
        point_b : array-like, shape=[..., 1]
            Point.

        Returns
        -------
        squared_dist : array-like, shape=[...]
            Riemannian squared distance.
        """
        ratio = gs.squeeze(point_b / point_a, axis=-1)
        return (gs.log(ratio)) ** 2

    def dist(self, point_a, point_b):
        """Compute the positive reals distance.

        Compute the Riemannian distance between point_a and point_b.

        Parameters
        ----------
        point_a : array-like, shape=[..., 1]
            Point.
        point_b : array-like, shape=[..., 1]
            Point.

        Returns
        -------
        dist : array-like, shape=[...]
            Riemannian distance.
        """
        ratio = gs.squeeze(point_b / point_a, axis=-1)
        return gs.abs(gs.log(ratio))
